### import#####################################################################


import collections
import pathlib

from PySide2 import QtWidgets, QtCore

import pyqtgraph as pg

from yaqc_cmds.project import project_globals as g

colors = g.colors_dict.read()
__here__ = pathlib.Path(__file__).parent


### basic elements ############################################################


class ExpandingWidget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.setMinimumHeight(0)
        self.setMinimumWidth(0)
        self.layout().setStretchFactor(self, 1)

    def sizeHint(self):
        return QtCore.QSize(16777215, 16777215)

    def add_to_layout(self, layout):
        layout.addWidget(self)
        layout.setStretchFactor(self, 16777215)


class Line(QtWidgets.QFrame):
    """
    direction: 'V' or 'H'
    """

    def __init__(self, direction):
        QtWidgets.QFrame.__init__(self)
        if direction == "V":
            self.setFrameShape(QtWidgets.QFrame.VLine)
        else:
            self.setFrameShape(QtWidgets.QFrame.HLine)
        StyleSheet = (
            "QFrame{border: 2px solid custom_color; border-radius: 0px; padding: 0px;}".replace(
                "custom_color", colors["widget_background"]
            )
        )
        self.setStyleSheet(StyleSheet)


line = Line  # legacy


class scroll_area(QtWidgets.QScrollArea):
    def __init__(self, show_bar=True):
        QtWidgets.QScrollArea.__init__(self)
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setWidgetResizable(True)
        if show_bar:
            self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        StyleSheet = "QScrollArea, QWidget{background: custom_color;}".replace(
            "custom_color", colors["background"]
        )
        StyleSheet += "QScrollBar{background: custom_color;}".replace(
            "custom_color", colors["widget_background"]
        )
        self.setStyleSheet(StyleSheet)


class Led(QtWidgets.QCheckBox):
    def __init__(self):
        QtWidgets.QCheckBox.__init__(self)
        self.setDisabled(True)
        path = str(__here__).replace("\\", "/")
        StyleSheet = f"QCheckBox::indicator:checked {{image: url({path}/widget files/checkbox_checked.png);}}"
        StyleSheet += f"QCheckBox::indicator:unchecked {{image: url({path}/widget files/checkbox_unchecked.png);}}"
        print(StyleSheet)
        self.setStyleSheet(StyleSheet)


### general ###################################################################


class SpinboxAsDisplay(QtWidgets.QDoubleSpinBox):
    def __init__(self, font_size=14, decimals=6, justify="right"):
        QtWidgets.QDoubleSpinBox.__init__(self)
        self.setValue(0.0)
        self.setDisabled(True)
        self.decimals_input = decimals
        self.setDecimals(decimals)
        self.setMinimum(-100000)
        self.setMaximum(100000)
        if justify == "right":
            self.setAlignment(QtCore.Qt.AlignRight)
        else:
            self.setAlignment(QtCore.Qt.AlignLeft)
        self.setMinimumWidth(0)
        self.setMaximumWidth(600)
        self.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        StyleSheet = "QDoubleSpinBox{color: custom_color_1; font: bold font_sizepx; border: 0px solid #000000;}".replace(
            "custom_color_1", g.colors_dict.read()["text_light"]
        ).replace(
            "font_size", str(int(font_size))
        )
        StyleSheet += (
            "QScrollArea, QWidget{background: custom_color;  border-color: black;}".replace(
                "custom_color", g.colors_dict.read()["background"]
            )
        )
        self.setStyleSheet(StyleSheet)


class Shutdown_button(QtWidgets.QPushButton):
    shutdown_go = QtCore.Signal()

    def __init__(self):
        QtWidgets.QPushButton.__init__(self)
        self.clicked.connect(self.initiate_shutdown)
        self.setText("SHUT DOWN")
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["stop"]
        )
        self.setStyleSheet(StyleSheet)

    def initiate_shutdown(self):
        self.setText("SHUTTING DOWN")
        g.app.read().processEvents()
        self.shutdown_go.emit()


class InputTable(QtWidgets.QWidget):
    def __getitem__(self, key):
        return self._dict[key]

    def __init__(self, width=130):
        """
        width of 160 good for modules
        """
        QtWidgets.QWidget.__init__(self)
        self.width_input = width
        self.setLayout(QtWidgets.QGridLayout())
        self.layout().setColumnMinimumWidth(0, width)
        self.layout().setColumnMinimumWidth(1, width)
        self.layout().setMargin(0)
        self.row_number = 0
        self.controls = []
        self._dict = collections.OrderedDict()

    def add(self, name, global_object, key=None):
        if key is None:
            key = name
        if global_object is None:
            global_type = "heading"
        else:
            global_type = global_object.type
            self._dict[key] = global_object
        getattr(self, global_type)(name, global_object)

    def busy(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        if name in ["DAQ status", "Status"]:  # hardcoded exceptions
            StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
                "custom_color", colors["text_light"]
            )
        else:
            StyleSheet = "QLabel{color: custom_color; font: bold 14px;}".replace(
                "custom_color", colors["heading_0"]
            )
        heading.setStyleSheet(StyleSheet)
        self.layout().addWidget(heading, self.row_number, 0)
        # control
        control = QtWidgets.QCheckBox()
        local_busy_display = BusyDisplay(global_object, global_object.update_signal)
        # finish
        self.layout().addWidget(local_busy_display, self.row_number, 1)
        self.controls.append(control)
        self.row_number += 1

    def heading(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        StyleSheet = "QLabel{color: custom_color; font: bold 14px;}".replace(
            "custom_color", colors["heading_0"]
        )
        heading.setStyleSheet(StyleSheet)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        self.layout().addWidget(heading, self.row_number, 0)
        self.controls.append(None)
        self.row_number += 1

    def number(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
            "custom_color", colors["text_light"]
        )
        heading.setStyleSheet(StyleSheet)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        self.layout().addWidget(heading, self.row_number, 0)
        # layout
        container_widget = QtWidgets.QWidget()
        container_widget.setLayout(QtWidgets.QHBoxLayout())
        layout = container_widget.layout()
        layout.setMargin(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        # control
        control = QtWidgets.QDoubleSpinBox()
        if global_object.display:
            control.setDisabled(True)
            StyleSheet = "QDoubleSpinBox{color: custom_color_1; font: bold font_sizepx; border: 0px solid #000000;}".replace(
                "custom_color_1", g.colors_dict.read()["text_light"]
            ).replace(
                "font_size", str(int(14))
            )
            StyleSheet += (
                "QScrollArea, QWidget{background: custom_color;  border-color: black;}".replace(
                    "custom_color", g.colors_dict.read()["background"]
                )
            )
        else:
            StyleSheet = "QDoubleSpinBox{color: custom_color; font: 14px;}".replace(
                "custom_color", colors["text_light"]
            )
            StyleSheet += "QScrollArea, QWidget{color: custom_color_1; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += "QWidget:disabled{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_disabled"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
        control.setStyleSheet(StyleSheet)
        control.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        global_object.give_control(control)
        layout.addWidget(control)
        # units combobox
        if not global_object.units is None:
            control.setMinimumWidth(self.width_input - 55)
            control.setMaximumWidth(self.width_input - 55)
            units = QtWidgets.QComboBox()
            units.setMinimumWidth(50)
            units.setMaximumWidth(50)
            StyleSheet = "QComboBox{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += "QComboBox:disabled{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_disabled"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += (
                "QAbstractItemView{color: custom_color_1; font: 50px solid white;}".replace(
                    "custom_color_1", colors["text_light"]
                ).replace("custom_color_2", colors["widget_background"])
            )
            units.setStyleSheet(StyleSheet)
            layout.addWidget(units)
            global_object.give_units_combo(units)
        # finish
        self.layout().addWidget(container_widget, self.row_number, 1)
        self.controls.append(container_widget)
        self.row_number += 1

    def string(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
            "custom_color", colors["text_light"]
        )
        heading.setStyleSheet(StyleSheet)
        self.layout().addWidget(heading, self.row_number, 0)
        # control
        control = QtWidgets.QLineEdit()
        control.setMinimumWidth(self.width_input)
        control.setMaximumWidth(self.width_input)
        if global_object.display:
            control.setDisabled(True)
            StyleSheet = "QWidget{color: custom_color_1; font: bold 14px; border: 0px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
        else:
            StyleSheet = "QWidget{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += "QWidget:disabled{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_disabled"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
        control.setStyleSheet(StyleSheet)
        global_object.give_control(control)
        # finish
        self.layout().addWidget(control, self.row_number, 1)
        self.controls.append(control)
        self.row_number += 1

    def combo(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
            "custom_color", colors["text_light"]
        )
        heading.setStyleSheet(StyleSheet)
        self.layout().addWidget(heading, self.row_number, 0)
        # control
        control = QtWidgets.QComboBox()
        control.setMinimumWidth(self.width_input)
        control.setMaximumWidth(self.width_input)
        if global_object.display:
            control.setDisabled(True)
            StyleSheet = "QComboBox{color: custom_color_1; font: bold 14px; border: 0px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            # StyleSheet += 'QComboBox:disabled{color: custom_color_1; font: 14px; border: 0px solid custom_color_2; border-radius: 1px;}'.replace('custom_color_1', colors['text_disabled']).replace('custom_color_2', colors['widget_background'])
            StyleSheet += "QAbstractItemView{color: custom_color_1; font: 50px solid white; border: 0px white}".replace(
                "custom_color_1", colors["widget_background"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += "QComboBox::drop-down{border: 0;}"
        else:
            StyleSheet = "QComboBox{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_light"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += "QComboBox:disabled{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
                "custom_color_1", colors["text_disabled"]
            ).replace(
                "custom_color_2", colors["widget_background"]
            )
            StyleSheet += (
                "QAbstractItemView{color: custom_color_1; font: 50px solid white;}".replace(
                    "custom_color_1", colors["text_light"]
                ).replace("custom_color_2", colors["widget_background"])
            )
        control.setStyleSheet(StyleSheet)
        global_object.give_control(control)
        # finish
        self.layout().addWidget(control, self.row_number, 1)
        self.controls.append(control)
        self.row_number += 1

    def checkbox(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
            "custom_color", colors["text_light"]
        )
        heading.setStyleSheet(StyleSheet)
        self.layout().addWidget(heading, self.row_number, 0)
        # control
        if global_object.display:
            control = Led()
        else:
            control = QtWidgets.QCheckBox()
        global_object.give_control(control)
        # finish
        self.layout().addWidget(control, self.row_number, 1)
        self.controls.append(control)
        self.row_number += 1

    def filepath(self, name, global_object):
        # heading
        heading = QtWidgets.QLabel(name)
        heading.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        StyleSheet = "QLabel{color: custom_color; font: 14px;}".replace(
            "custom_color", colors["text_light"]
        )
        heading.setStyleSheet(StyleSheet)
        self.layout().addWidget(heading, self.row_number, 0)
        # layout
        container_widget = QtWidgets.QWidget()
        container_widget.setLayout(QtWidgets.QHBoxLayout())
        layout = container_widget.layout()
        layout.setMargin(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        # push button
        load_button = QtWidgets.QPushButton("Load")
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["go"]
        )
        load_button.setStyleSheet(StyleSheet)
        load_button.setMinimumHeight(20)
        load_button.setMaximumHeight(20)
        load_button.setMinimumWidth(40)
        load_button.setMaximumWidth(40)
        layout.addWidget(load_button)
        global_object.give_button(load_button)
        # display
        display = QtWidgets.QLineEdit()
        # display.setDisabled(True)
        display.setReadOnly(True)
        display.setMinimumWidth(self.width_input - 45)
        display.setMaximumWidth(self.width_input - 45)
        StyleSheet = "QWidget{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
            "custom_color_1", colors["text_light"]
        ).replace(
            "custom_color_2", colors["widget_background"]
        )
        StyleSheet += "QWidget:disabled{color: custom_color_1; font: 14px; border: 1px solid custom_color_2; border-radius: 1px;}".replace(
            "custom_color_1", colors["text_disabled"]
        ).replace(
            "custom_color_2", colors["widget_background"]
        )
        display.setStyleSheet(StyleSheet)
        layout.addWidget(display)
        global_object.give_control(display)
        # finish
        self.layout().addWidget(container_widget, self.row_number, 1)
        self.controls.append(container_widget)
        self.row_number += 1


class Label(QtWidgets.QLabel):
    def __init__(self, text, color="text_light", bold=False, font_size=14):
        QtWidgets.QLabel.__init__(self, text)
        if bold:
            bold_status = "bold"
        else:
            bold_status = ""
        StyleSheet = f"QLabel{{color: {colors[color]}; font: {font_size}px;}}".replace(
            "bold_status", bold_status
        )
        self.setStyleSheet(StyleSheet)


class SetButton(QtWidgets.QPushButton):
    def __init__(self, text, color="go"):
        QtWidgets.QPushButton.__init__(self)
        self.setText(text)
        self.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors[color]
        )
        self.setStyleSheet(StyleSheet)


class TableWidget(QtWidgets.QTableWidget):
    def __init__(self):
        QtWidgets.QTableWidget.__init__(self)
        StyleSheet = "QTableWidget::item{padding: 0px}"
        StyleSheet += "QHeaderView::section{background: background_color; color:white; font: bold 14px}".replace(
            "background_color", colors["background"]
        )
        StyleSheet += "QTableWidget{background-color: custom_color;}".replace(
            "custom_color", colors["background"]
        )
        self.setStyleSheet(StyleSheet)


class TabWidget(QtWidgets.QTabWidget):
    def __init__(self):
        QtWidgets.QTabWidget.__init__(self)
        StyleSheet = "QTabBar::tab{width: 130px;}"
        self.setStyleSheet(StyleSheet)


### hardware ##################################################################


class BusyDisplay(QtWidgets.QPushButton):
    """
    access value object to get state

    True: scan running

    pause methods are built in to this object
    """

    def __init__(self, busy_object, update_signal):
        QtWidgets.QPushButton.__init__(self)
        self.busy_object = busy_object
        update_signal.connect(self.update)
        self.setText("READY")
        self.setMinimumHeight(15)
        self.setMaximumHeight(15)
        StyleSheet = "QPushButton{background:background_color; border-width:0px;  border-radius: 0px; font: bold 14px; color: text_color}".replace(
            "background_color", colors["background"]
        ).replace(
            "text_color", colors["background"]
        )
        self.setStyleSheet(StyleSheet)
        self.update()

    def update(self):
        if self.busy_object.read():
            self.setText("BUSY")
            StyleSheet = "QPushButton{background:background_color; border-width:0px;  border-radius: 0px; font: bold 14px; color: text_color; text-align: left}".replace(
                "background_color", colors["background"]
            ).replace(
                "text_color", colors["stop"]
            )
            self.setStyleSheet(StyleSheet)
        else:
            self.setText("READY")
            StyleSheet = "QPushButton{background:background_color; border-width:0px;  border-radius: 0px; font: bold 14px; color: text_color}".replace(
                "background_color", colors["background"]
            ).replace(
                "text_color", colors["background"]
            )
            self.setStyleSheet(StyleSheet)


class Hardware_control_table(QtWidgets.QWidget):
    def __init__(self, hardware_names, combobox=False, combobox_label=""):
        QtWidgets.QWidget.__init__(self)
        self.setLayout(QtWidgets.QGridLayout())
        self.layout().setMargin(0)
        self.comboboxes = []
        self.current_position_displays = []
        self.new_position_controls = []
        self.set_buttons = []
        for i in range(len(hardware_names)):
            # names
            collumn = 0
            label = QtWidgets.QLabel(hardware_names[i])
            StyleSheet = "QLabel{color: custom_color; font: bold 14px;}".replace(
                "custom_color", colors["heading_0"]
            )
            label.setStyleSheet(StyleSheet)
            self.layout().addWidget(label, i, collumn)
            # comboboxes
            if combobox:
                collumn += 1
                combobox_obj = QtWidgets.QComboBox()
                self.layout().addWidget(combobox_obj, i, collumn)
                self.comboboxes.append(combobox_obj)
            # current
            collumn += 1
            current_position = QtWidgets.QDoubleSpinBox()
            current_position.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
            current_position.setDisabled(True)
            self.layout().addWidget(current_position, i, collumn)
            self.current_position_displays.append(current_position)
            # new
            collumn += 1
            new_position = QtWidgets.QDoubleSpinBox()
            new_position.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
            self.layout().addWidget(new_position, i, collumn)
            self.new_position_controls.append(new_position)
            self.set_allowed_values(i, 0, 10000, 2, 10)

    def set_allowed_values(self, index, min_val, max_val, decimals=2, single_step=1.00):
        controls = [
            self.current_position_displays[index],
            self.new_position_controls[index],
        ]
        for control in controls:
            control.setMinimum(min_val)
            control.setMaximum(max_val)
            control.setDecimals(decimals)
            control.setSingleStep(single_step)


class HardwareLayoutWidget(QtWidgets.QGroupBox):
    def __init__(self, name):
        QtWidgets.QGroupBox.__init__(self)
        # layout
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setMargin(5)
        # container
        heading_container = QtWidgets.QWidget()
        heading_container.setLayout(QtWidgets.QHBoxLayout())
        heading_container.layout().setMargin(0)
        # add heading
        heading = QtWidgets.QLabel(name)
        StyleSheet = "QLabel{color: custom_color; font: bold 14px;}".replace(
            "custom_color", colors["heading_1"]
        )
        heading.setStyleSheet(StyleSheet)
        heading_container.layout().addWidget(heading)
        self.layout().addWidget(heading_container)

    def add_buttons(self, set_method, advanced_method, hardwares=[]):
        # layout
        button_container = QtWidgets.QWidget()
        button_container.setLayout(QtWidgets.QHBoxLayout())
        button_container.layout().setMargin(0)
        # advanced
        advanced_button = QtWidgets.QPushButton()
        advanced_button.setText("ADVANCED")
        advanced_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["advanced"]
        )
        advanced_button.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(advanced_button)
        advanced_button.clicked.connect(advanced_method)
        # set
        set_button = QtWidgets.QPushButton()
        set_button.setText("SET")
        set_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["set"]
        )
        set_button.setStyleSheet(StyleSheet)
        # add
        self.layout().addWidget(button_container)
        # connect to signals
        g.queue_control.disable_when_true(set_button)
        button_container.layout().addWidget(set_button)
        set_button.clicked.connect(set_method)
        for hardware in hardwares:
            set_button.setDisabled(hardware.busy.read())  # first time
            hardware.update_ui.connect(lambda: set_button_decide(set_button, hardwares))
        return [advanced_button, set_button]


def set_button_decide(set_button, hardwares):
    if g.queue_control.read():
        set_button.setDisabled(True)
    else:
        set_button.setDisabled(False)
        for hardware in hardwares:
            if hardware.busy.read():
                set_button.setDisabled(hardware.busy.read())


class HardwareFrontPanel(QtCore.QObject):
    advanced = QtCore.Signal()

    def __init__(self, hardwares, name="Hardware"):
        QtCore.QObject.__init__(self)
        self.name = name
        # link hardware object signals
        self.hardwares = hardwares
        for hardware in self.hardwares:
            hardware.update_ui.connect(self.update)
        # create gui
        self.create_frame()

    def create_frame(self):
        layout_widget = HardwareLayoutWidget(self.name)
        layout = layout_widget.layout()
        # layout table
        input_table = InputTable(130)
        self.front_panel_elements = []
        for hardware in self.hardwares:
            name = " ".join([hardware.name, "(" + hardware.model + ")"])
            input_table.add(name, hardware.busy)
            current_objects = hardware.exposed
            destination_objects = []
            for obj in hardware.exposed:
                input_table.add(obj.label, obj)
                dest_obj = obj.associate(display=False, pre_name="Dest. ")
                destination_objects.append(dest_obj)
                if hasattr(obj, "units_updated"):
                    obj.units_updated.connect(self.on_position_units_updated)
                    dest_obj.units_updated.connect(self.on_destination_units_updated)
            for obj in destination_objects:
                input_table.add(obj.label, obj)
            self.front_panel_elements.append([current_objects, destination_objects])
            hardware.initialized.updated.connect(self.initialize)
        layout.addWidget(input_table)
        self.advanced_button, self.set_button = layout_widget.add_buttons(
            self.on_set, self.show_advanced, self.hardwares
        )
        g.hardware_widget.add_to(layout_widget)

    def initialize(self):
        # will fire each time ANY hardware contained within finishes initialize
        # not ideal behavior, but good enough
        for hardware, front_panel_elements in zip(self.hardwares, self.front_panel_elements):
            if hardware.initialized:
                for current_object, destination_object in zip(
                    front_panel_elements[0], front_panel_elements[1]
                ):
                    position = current_object.read()
                    destination_object.write(position)

    def update(self):
        pass

    def show_advanced(self):
        self.advanced.emit()

    def on_destination_units_updated(self):
        for pl, dl in self.front_panel_elements:
            if hasattr(pl[0], "units"):
                pl[0].set_units(dl[0].units)

    def on_position_units_updated(self):
        for pl, dl in self.front_panel_elements:
            if hasattr(pl[0], "units"):
                dl[0].set_units(pl[0].units)

    def on_set(self):
        for hardware, front_panel_elements in zip(self.hardwares, self.front_panel_elements):
            for current_object, destination_object in zip(
                front_panel_elements[0], front_panel_elements[1]
            ):
                if current_object.set_method == "set_position":
                    hardware.set_position(
                        destination_object.read(),
                        destination_object.units,
                        force_send=True,
                    )
                else:
                    hardware.q.push(current_object.set_method, [destination_object.read()])
        g.coset_control.read().launch()

    def stop(self):
        pass


hardware_advanced_panels = []


class HardwareAdvancedPanel(QtCore.QObject):
    def __init__(self, hardwares, advanced_button):
        QtCore.QObject.__init__(self)
        self.tabs = QtWidgets.QTabWidget()
        for hardware in hardwares:
            widget = QtWidgets.QWidget()
            box = QtWidgets.QHBoxLayout()
            hardware.gui.create_frame(box)
            widget.setLayout(box)
            self.tabs.addTab(widget, hardware.name)
        # box
        self.box = QtWidgets.QVBoxLayout()
        hardware_advanced_box = g.hardware_advanced_box.read()
        hardware_advanced_box.addWidget(self.tabs)
        # link into advanced button press
        self.advanced_button = advanced_button
        self.advanced_button.clicked.connect(self.on_advanced)
        main_window = g.main_window.read()
        self.advanced_button.clicked.connect(lambda: main_window.tabs.setCurrentIndex(0))
        hardware_advanced_panels.append(self)
        self.hide()

    def hide(self):
        self.tabs.hide()
        self.advanced_button.setDisabled(False)

    def on_advanced(self):
        for panel in hardware_advanced_panels:
            panel.hide()
        self.tabs.show()
        # self.advanced_button.setDisabled(True)


### queue #####################################################################


class QueueControl(QtWidgets.QPushButton):
    launch_scan = QtCore.Signal()
    stop_scan = QtCore.Signal()

    def __init__(self):
        QtWidgets.QPushButton.__init__(self)
        self.clicked.connect(self.update)
        self.setMinimumHeight(25)
        self.set_style("RUN QUEUE", "go")
        self.value = False

    def set_style(self, text, color):
        self.setText(text)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors[color]
        )
        self.setStyleSheet(StyleSheet)


class ChoiceWindow(QtWidgets.QMessageBox):
    def __init__(self, title, button_labels):
        QtWidgets.QMessageBox.__init__(self)
        self.setWindowTitle(title)
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        for label in button_labels:
            self.addButton(label, QtWidgets.QMessageBox.YesRole)
        self.setIcon(self.NoIcon)

    def set_informative_text(self, text):
        self.setInformativeText(text)

    def set_text(self, text):
        self.setText(text)

    def show(self):
        """
        Returns the index of the chosen button
        """
        self.isActiveWindow()
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        return self.exec_()


### plotting ##################################################################


class Plot1D(pg.GraphicsView):
    def __init__(self, title=None, xAutoRange=True, yAutoRange=True):
        pg.GraphicsView.__init__(self)
        # create layout
        self.graphics_layout = pg.GraphicsLayout(border="w")
        self.setCentralItem(self.graphics_layout)
        self.graphics_layout.layout.setSpacing(0)
        self.graphics_layout.setContentsMargins(0.0, 0.0, 1.0, 1.0)
        # create plot object
        self.plot_object = self.graphics_layout.addPlot(0, 0)
        self.labelStyle = {"color": "#FFF", "font-size": "14px"}
        self.x_axis = self.plot_object.getAxis("bottom")
        self.x_axis.setLabel(**self.labelStyle)
        self.y_axis = self.plot_object.getAxis("left")
        self.y_axis.setLabel(**self.labelStyle)
        self.plot_object.showGrid(x=True, y=True, alpha=0.5)
        self.plot_object.setMouseEnabled(False, True)
        self.plot_object.enableAutoRange(x=xAutoRange, y=yAutoRange)
        # title
        if title:
            self.plot_object.setTitle(title)

    def add_scatter(self, color="c", size=3, symbol="o"):
        curve = pg.ScatterPlotItem(symbol=symbol, pen=(color), brush=(color), size=size)
        self.plot_object.addItem(curve)
        return curve

    def add_line(self, color="c", size=3, symbol="o"):
        curve = pg.PlotCurveItem(symbol=symbol, pen=(color), brush=(color), size=size)
        self.plot_object.addItem(curve)
        return curve

    def add_infinite_line(self, color="y", style="solid", angle=90.0, movable=False, hide=True):
        """
        Add an InfiniteLine object.

        Parameters
        ----------
        color : (optional)
            The color of the line. Accepts any argument valid for `pyqtgraph.mkColor <http://www.pyqtgraph.org/documentation/functions.html#pyqtgraph.mkColor>`_. Default is 'y', yellow.
        style : {'solid', 'dashed', dotted'} (optional)
            Linestyle. Default is solid.
        angle : float (optional)
            The angle of the line. 90 is vertical and 0 is horizontal. 90 is default.
        movable : bool (optional)
            Toggles if user can move the line. Default is False.
        hide : bool (optional)
            Toggles if the line is hidden upon initialization. Default is True.

        Returns
        -------
        InfiniteLine object
            Useful methods: setValue, show, hide
        """
        if style == "solid":
            linestyle = QtCore.Qt.SolidLine
        elif style == "dashed":
            linestyle = QtCore.Qt.DashLine
        elif style == "dotted":
            linestyle = QtCore.Qt.DotLine
        else:
            print("style not recognized in add_infinite_line")
            linestyle = QtCore.Qt.SolidLine
        pen = pg.mkPen(color, style=linestyle)
        line = pg.InfiniteLine(pen=pen)
        line.setAngle(angle)
        line.setMovable(movable)
        if hide:
            line.hide()
        self.plot_object.addItem(line)
        return line

    def set_labels(self, xlabel=None, ylabel=None):
        if xlabel:
            self.plot_object.setLabel("bottom", text=xlabel)
            self.plot_object.showLabel("bottom")
        if ylabel:
            self.plot_object.setLabel("left", text=ylabel)
            self.plot_object.showLabel("left")

    def set_xlim(self, xmin, xmax):
        self.plot_object.setXRange(xmin, xmax)

    def set_ylim(self, ymin, ymax):
        self.plot_object.setYRange(ymin, ymax)

    def clear(self):
        self.plot_object.clear()



"""
Parent hardware class and associated.
"""
__all__ = ["Driver", "GUI", "Hardware", "import_hardwares"]


### import ####################################################################


import pathlib
import time
import collections

from PySide2 import QtCore
from PySide2 import QtWidgets

import appdirs
import toml

import WrightTools as wt

from yaqc_cmds.project import classes as pc
from yaqc_cmds.project import widgets as pw
from yaqc_cmds.project import project_globals as g


__here__ = pathlib.Path(__file__)


### driver ####################################################################


class Driver(pc.Driver):
    initialized_signal = QtCore.Signal()

    def __init__(self, hardware, **kwargs):
        pc.Driver.__init__(self)
        # basic attributes
        self.hardware = hardware
        self.enqueued = self.hardware.enqueued
        self.busy = self.hardware.busy
        self.name = self.hardware.name
        self.model = self.hardware.model
        self.serial = self.hardware.serial
        self.label = pc.String(kwargs["label"], display=True)
        self.native_units = kwargs["native_units"]
        self.state_filepath = (
            pathlib.Path(appdirs.user_data_dir("yaqc-cmds", "yaqc-cmds"))
            / "hardware"
            / f"{self.name}-state.toml"
        )
        self.state_filepath.parent.mkdir(parents=True, exist_ok=True)
        # mutex attributes
        self.limits = pc.NumberLimits(units=self.native_units)
        if self.state_filepath.exists():
            state = toml.load(self.state_filepath)
        else:
            state = {}
        self.offset = pc.Number(
            initial_value=0, units=self.native_units, name="Offset", display=True
        )
        self.load_state(state)
        # attributes for 'exposure'
        self.exposed = [self.position]
        self.recorded = collections.OrderedDict()
        print(self.name, type(self.name))
        self.recorded[self.name] = [
            self.position,
            self.native_units,
            1.0,
            self.label.read(),
            False,
        ]
        # self.queue_emptied.connect(self.save_status)

    def close(self):
        pass

    def get_position(self):
        self.update_ui.emit()

    def initialize(self):
        """
        May not accept arguments.
        """
        print("Driver initialize", self.name)
        self.label.updated.connect(self.on_label_updated)
        self.initialized.write(True)
        self.initialized_signal.emit()

    @QtCore.Slot()
    def on_label_updated(self):
        self.recorded[self.name] = [
            self.position,
            self.native_units,
            1.0,
            self.label.read(),
            False,
        ]

    def poll(self):
        """
        polling only gets enqueued by Hardware when not in module control
        """
        self.get_position()
        self.is_busy()

    def get_state(self):
        return {
            "position": float(self.position.read(self.native_units)),
            "display_units": self.position.units,
            "offset": float(self.offset.read()),
        }

    def save_status(self):
        print(self.name, "STATE", self.get_state())
        with open(self.state_filepath, "w") as f:
            toml.dump(self.get_state(), f)

    def load_state(self, state):
        self.position = pc.Number(
            initial_value=state.get("position", 0),
            units=self.native_units,
            name="Position",
            display=True,
            set_method="set_position",
            limits=self.limits,
        )
        self.position.set_units(state.get("display_units", self.native_units))
        self.offset.write(state.get("offset", 0))

    def set_offset(self, offset):
        self.offset.write(offset, self.native_units)

    def set_position(self, destination):
        time.sleep(0.01)  # rate limiter for virtual hardware behavior
        self.position.write(destination, self.native_units)
        self.get_position()
        self.save_status()

    def wait_until_still(self):
        while self.is_busy():
            self.get_position()
            time.sleep(0.01)
        self.get_position()


### gui #######################################################################


class GUI(QtCore.QObject):
    def __init__(self, hardware):
        """
        Runs after driver.__init__, but before driver.initialize.
        """
        QtCore.QObject.__init__(self)
        self.hardware = hardware
        self.driver = hardware.driver

    def close(self):
        pass

    def create_frame(self, layout):
        """
        Runs before initialize.
        """
        # layout
        layout.setMargin(5)
        self.layout = layout
        # scroll area
        scroll_container_widget = QtWidgets.QWidget()
        self.scroll_area = pw.scroll_area(show_bar=False)
        self.scroll_area.setWidget(scroll_container_widget)
        self.scroll_area.setMinimumWidth(300)
        self.scroll_area.setMaximumWidth(300)
        scroll_container_widget.setLayout(QtWidgets.QVBoxLayout())
        self.scroll_layout = scroll_container_widget.layout()
        self.scroll_layout.setMargin(5)
        # attributes table
        self.attributes_table = pw.InputTable()
        self.attributes_table.add("Attributes", None)
        name = pc.String(self.hardware.name, display=True)
        self.attributes_table.add("Name", name)
        model = pc.String(self.hardware.model, display=True)
        self.attributes_table.add("Model", model)
        serial = pc.String(self.hardware.serial, display=True)
        self.attributes_table.add("Serial", serial)
        self.position = self.hardware.position.associate()
        self.hardware.position.updated.connect(self.on_position_updated)
        self.attributes_table.add("Label", self.hardware.driver.label)
        self.attributes_table.add("Position", self.position)
        self.offset = self.hardware.offset.associate()
        self.hardware.offset.updated.connect(self.on_offset_updated)
        self.attributes_table.add("Offset", self.offset)
        # initialization
        if self.hardware.initialized.read():
            self.initialize()
        else:
            self.hardware.initialized_signal.connect(self.initialize)

    def initialize(self):
        """
        Runs only once the hardware is done initializing.
        """
        self.layout.addWidget(self.scroll_area)
        # attributes
        self.scroll_layout.addWidget(self.attributes_table)
        # stretch
        self.scroll_layout.addStretch(1)
        self.layout.addStretch(1)

    def on_offset_updated(self):
        new = self.hardware.offset.read(self.hardware.native_units)
        self.offset.write(new, self.hardware.native_units)

    def on_position_updated(self):
        new = self.hardware.position.read(self.hardware.native_units)
        self.position.write(new, self.hardware.native_units)


### hardware ##################################################################


hardwares = []


def all_initialized():
    # fires any time a hardware is initialized
    for hardware in hardwares:
        while not hardware.initialized.read():
            print(hardware.name, "not yet inited")
            time.sleep(0.1)
    # past here only runs when ALL hardwares are initialized
    g.hardware_initialized.write(True)


class Hardware(pc.Hardware):
    def __init__(self, *args, **kwargs):
        pc.Hardware.__init__(self, *args, **kwargs)
        self.driver.initialized_signal.connect(self.on_address_initialized)
        self.exposed = self.driver.exposed
        for obj in self.exposed:
            obj.updated.connect(self.update)
        self.recorded = self.driver.recorded
        self.offset = self.driver.offset
        self.position = self.exposed[0]
        self.native_units = self.driver.native_units
        self.destination = pc.Number(units=self.native_units, display=True)
        self.destination.write(self.position.read(self.native_units), self.native_units)
        self.limits = self.driver.limits
        hardwares.append(self)

    def close(self):
        self.q.push("save_status")
        pc.Hardware.close(self)

    def get_destination(self, output_units="same"):
        print("get_destination", self.name, self.destination.read(output_units))
        return self.destination.read(output_units=output_units)

    def get_position(self, output_units="same"):
        return self.position.read(output_units=output_units)

    def is_valid(self, destination, input_units=None):
        if input_units is None:
            pass
        else:
            destination = wt.units.converter(destination, input_units, self.native_units)
        min_value, max_value = self.limits.read(self.native_units)
        if min_value <= destination <= max_value:
            return True
        else:
            return False

    def on_address_initialized(self):
        self.destination.write(self.get_position(), self.native_units)
        # all_initialized()
        self.initialized_signal.emit()

    def poll(self, force=False):
        if force:
            self.q.push("poll")
            self.get_position()
        elif not g.queue_control.read():
            self.q.push("poll")
            self.get_position()

    def set_offset(self, offset, input_units=None):
        if input_units is None:
            pass
        else:
            offset = wt.units.converter(offset, input_units, self.native_units)
        # do nothing if new offset is same as current offset
        if offset == self.offset.read(self.native_units):
            return
        self.q.push("set_offset", offset)

    def set_position(self, destination, input_units=None, force_send=False):
        if input_units is None:
            pass
        else:
            destination = wt.units.converter(destination, input_units, self.native_units)
        # do nothing if new destination is same as current destination
        if destination == self.destination.read(self.native_units):
            if not force_send:
                return
        self.destination.write(destination, self.native_units)
        self.q.push("set_position", destination)

    @property
    def units(self):
        return self.position.units


### import method #############################################################


def import_hardwares(config, name, Driver, GUI, Hardware):
    hardwares = []
    for hw_name, section in config.items():
        if section.get("enable", True):
            # initialization arguments
            kwargs = collections.OrderedDict()
            kwargs.update(section)
            for option in ["__name__", "enable", "model", "serial", "path"]:
                kwargs.pop(option, None)
            model = section.get("model", "yaq")
            hardware = Hardware(Driver, kwargs, GUI, name=hw_name, model=model)
            hardwares.append(hardware)
    gui = pw.HardwareFrontPanel(hardwares, name=name)
    advanced_gui = pw.HardwareAdvancedPanel(hardwares, gui.advanced_button)
    return hardwares, gui, advanced_gui



#!/usr/bin/env python

import sys

sys.dont_write_bytecode = True  # ??? - Blaise 2020-08-13

import glob
import sys
import time
import logging
from argparse import ArgumentParser

from slackclient import SlackClient

import appdirs
import toml
import pathlib

config = toml.load(pathlib.Path(appdirs.user_config_dir("yaqc-cmds", "yaqc-cmds")) / "config.toml")
default_channel = config["slack"]["channel"]
witch_token = config["slack"]["token"]


class RtmBot(object):
    def __init__(self, token):
        self.last_ping = 0
        self.token = token
        self.bot_plugins = []
        self.slack_client = None

    def connect(self):
        """Convenience method that creates Server instance"""
        self.slack_client = SlackClient(self.token)
        self.slack_client.rtm_connect()

    def start(self):
        self.connect()
        self.load_plugins()

    def run_plugins(self):
        for reply in self.rtm_read():
            self.input(reply)
        self.crons()
        self.output()
        self.autoping()

    def rtm_read(self):
        return self.slack_client.rtm_read()

    def autoping(self):
        # hardcode the interval to 3 seconds
        now = int(time.time())
        if now > self.last_ping + 3:
            self.slack_client.server.ping()
            self.last_ping = now

    def input(self, data):
        if "type" in data:
            function_name = "process_" + data["type"]
            for plugin in self.bot_plugins:
                plugin.register_jobs()
                plugin.do(function_name, data)

    def output(self):
        for plugin in self.bot_plugins:
            limiter = False
            for output in plugin.do_output():
                channel = self.slack_client.server.channels.find(output[0])
                if channel != None and output[1] != None:
                    if limiter == True:
                        time.sleep(0.1)
                        limiter = False
                    message = output[1].encode("ascii", "ignore")
                    channel.send_message("{}".format(message))
                    limiter = True

    def crons(self):
        for plugin in self.bot_plugins:
            plugin.do_jobs()

    def load_plugins(self):
        for plugin in glob.glob(directory + "/plugins/*"):
            sys.path.insert(0, plugin)
            sys.path.insert(0, directory + "/plugins/")
        for plugin in glob.glob(directory + "/plugins/*.py") + glob.glob(
            directory + "/plugins/*/*.py"
        ):
            logging.info(plugin)
            name = plugin.split("/")[-1][:-3]
            self.bot_plugins.append(Plugin(name))


class Plugin(object):
    def __init__(self, name, plugin_config={}):
        self.name = name
        self.jobs = []
        self.module = __import__(name)
        self.register_jobs()
        self.outputs = []
        if name in config:
            logging.info("config found for: " + name)
            self.module.config = config[name]
        if "setup" in dir(self.module):
            self.module.setup()

    def register_jobs(self):
        if "crontable" in dir(self.module):
            for interval, function in self.module.crontable:
                self.jobs.append(Job(interval, eval("self.module." + function)))
            logging.info(self.module.crontable)
            self.module.crontable = []
        else:
            self.module.crontable = []

    def do(self, function_name, data):
        if function_name in dir(self.module):
            eval("self.module." + function_name)(data)
        if "catch_all" in dir(self.module):
            self.module.catch_all(data)

    def do_jobs(self):
        for job in self.jobs:
            job.check()

    def do_output(self):
        output = []
        while True:
            if "outputs" in dir(self.module):
                if len(self.module.outputs) > 0:
                    logging.info("output from {}".format(self.module))
                    output.append(self.module.outputs.pop(0))
                else:
                    break
            else:
                self.module.outputs = []
        return output


class Job(object):
    def __init__(self, interval, function):
        self.function = function
        self.interval = interval
        self.lastrun = 0

    def __str__(self):
        return "{} {} {}".format(self.function, self.interval, self.lastrun)

    def __repr__(self):
        return self.__str__()

    def check(self):
        if self.lastrun + self.interval < time.time():
            self.function()
            self.lastrun = time.time()


class UnknownChannel(Exception):
    pass


def main_loop():
    if "LOGFILE" in config:
        logging.basicConfig(
            filename=config["LOGFILE"],
            level=logging.INFO,
            format="%(asctime)s %(message)s",
        )
    logging.info(directory)
    try:
        bot.start()
    except KeyboardInterrupt:
        sys.exit(0)
    except:
        logging.exception("OOPS")


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("-c", "--config", help="Full path to config file.", metavar="path")
    return parser.parse_args()



# --- import --------------------------------------------------------------------------------------


import pathlib
import time

import toml
import appdirs

import WrightTools as wt
import yaqc

import yaqc_cmds.__main__
import yaqc_cmds.project.project_globals as g
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.classes as pc
import yaqc_cmds.hardware.hardware as hw

# --- driver --------------------------------------------------------------------------------------


class Driver(hw.Driver):
    def __init__(self, *args, **kwargs):
        self.yaqd_port = kwargs["yaqd_port"]
        self.yaqd_host = kwargs.get("yaqd_host", "127.0.0.1")
        self.motor = yaqc.Client(self.yaqd_port, host=self.yaqd_host)
        self.motor_units = self.motor.get_units()
        self.native_units = kwargs.get("native_units", "deg")
        self.native_per_motor = float(wt.units.convert(1, self.motor_units, self.native_units))
        hw.Driver.__init__(self, *args, **kwargs)
        id_ = self.motor.id()
        if id_["model"] is not None:
            self.hardware.model = id_["model"]
        elif id_["kind"].startswith("fake"):
            self.hardware.model = "fake"
        else:
            self.hardware.model = id_["kind"]

        self.factor = self.hardware.factor
        self.factor.write(kwargs["factor"])
        self.motor_limits = self.hardware.motor_limits
        self.motor_limits.write(*self.motor.get_limits())
        self.motor_position = self.hardware.motor_position
        self.zero_position = self.hardware.zero_position
        self.recorded["_".join([self.name, "zero"])] = [
            self.zero_position,
            "deg",
            0,
            self.label.read(),
            True,
        ]

    def initialize(self):
        # This should be unnecessary at some point, once everything is yaq
        self.get_position()
        self.initialized.write(True)
        self.initialized_signal.emit()

    def home(self):
        self.motor.home()
        self.wait_until_still()

    def is_busy(self):
        return self.motor.busy()

    def get_position(self):
        position = self.motor.get_position()
        self.motor_position.write(position)
        filter = (
            (position - self.zero_position.read(self.motor_units))
            * self.native_per_motor
            * self.factor.read()
        )
        self.position.write(filter, self.native_units)
        return filter

    def get_state(self):
        state = super().get_state()
        state["zero_position"] = float(self.zero_position.read(self.motor_units))
        return state

    def load_state(self, state):
        super().load_state(state)
        # called before self.zero_position aliases the hardware one
        self.hardware.zero_position.write(state.get("zero_position", 0))

    def set_motor_position(self, motor_position):
        self.motor.set_position(motor_position)
        time.sleep(0.01)
        self.wait_until_still()
        self.get_position()

    def set_position(self, destination):
        destination_motor = self.zero_position.read(self.motor_units) + destination / (
            self.native_per_motor * self.factor.read()
        )
        self.set_motor_position(destination_motor)

    def set_offset(self, offset):
        # update zero
        offset_from_here = offset - self.offset.read(self.native_units)
        offset_motor = offset_from_here / (self.native_per_motor * self.factor.read())
        new_zero = self.zero_position.read(self.motor_units) + offset_motor
        self.set_zero(new_zero)
        self.offset.write(offset, self.native_units)
        # return to old position
        destination = self.hardware.destination.read(self.native_units)
        self.set_position(destination)

    def update_recorded(self):
        self.recorded.clear()
        self.recorded[self.name] = [
            self.position,
            self.native_units,
            1.0,
            self.label.read(),
            False,
        ]
        self.recorded[f"{self.name}_position"] = [
            self.motor_position,
            self.motor_units,
            1.0,
            self.label.read(),
            False,
        ]
        self.recorded[f"{self.name}_zero"] = [
            self.zero_position,
            self.motor_units,
            1.0,
            self.label.read(),
            False,
        ]

    def set_zero(self, new_zero):
        self.zero_position.write(new_zero)
        motor_min, motor_max = self.motor_limits.read()
        min_ = (motor_min - new_zero) * self.native_per_motor * self.factor.read()
        max_ = (motor_max - new_zero) * self.native_per_motor * self.factor.read()
        self.limits.write(min_, max_, self.native_units)
        self.get_position()


# --- gui -----------------------------------------------------------------------------------------


class GUI(hw.GUI):
    def initialize(self):
        self.layout.addWidget(self.scroll_area)
        # attributes
        self.attributes_table.add("Label", self.hardware.label)
        self.attributes_table.add("Factor", self.hardware.factor)
        self.scroll_layout.addWidget(self.attributes_table)
        # deg input table
        input_table = pw.InputTable()
        input_table.add("Motor Position", None)
        input_table.add("Current", self.hardware.motor_position)
        self.motor_destination = self.hardware.motor_position.associate(display=False)
        input_table.add("Destination", self.motor_destination)
        self.scroll_layout.addWidget(input_table)
        # set deg button
        self.set_motor_button = pw.SetButton("SET POSITION")
        self.scroll_layout.addWidget(self.set_motor_button)
        self.set_motor_button.clicked.connect(self.on_set_motor)
        g.queue_control.disable_when_true(self.set_motor_button)
        # zero input table
        input_table = pw.InputTable()
        input_table.add("Zero Position", None)
        input_table.add("Current", self.hardware.zero_position)
        self.zero_destination = self.hardware.zero_position.associate(display=False)
        input_table.add("Destination", self.zero_destination)
        self.scroll_layout.addWidget(input_table)
        # set zero button
        self.set_zero_button = pw.SetButton("SET ZERO")
        self.scroll_layout.addWidget(self.set_zero_button)
        self.set_zero_button.clicked.connect(self.on_set_zero)
        g.queue_control.disable_when_true(self.set_zero_button)
        # horizontal line
        self.scroll_layout.addWidget(pw.line("H"))
        # home button
        self.home_button = pw.SetButton("HOME", "advanced")
        self.scroll_layout.addWidget(self.home_button)
        self.home_button.clicked.connect(self.on_home)
        g.queue_control.disable_when_true(self.home_button)
        # finish
        self.scroll_layout.addStretch(1)
        self.layout.addStretch(1)
        self.hardware.update_ui.connect(self.update)

    def on_home(self):
        self.driver.hardware.q.push("home")

    def on_set_motor(self):
        new_deg = self.motor_destination.read("deg")
        self.hardware.set_motor_position(new_deg, units="deg")

    def on_set_zero(self):
        new_zero = self.zero_destination.read("deg")
        print("on_set_new_zero", new_zero)
        self.driver.set_zero(new_zero)
        self.driver.offset.write(0)
        name = self.hardware.name
        g.coset_control.read().zero(name)
        self.driver.get_position()
        self.driver.save_status()

    def update(self):
        pass


# --- hardware ------------------------------------------------------------------------------------


class Hardware(hw.Hardware):
    def __init__(self, *arks, **kwargs):
        self.kind = "filter"
        self.factor = pc.Number(1, decimals=0, display=True)
        self.motor_limits = pc.NumberLimits(min_value=-360.0, max_value=360.0, units="deg")
        self.motor_position = pc.Number(units="deg", display=True, limits=self.motor_limits)
        self.zero_position = pc.Number(display=True, units="deg", limits=self.motor_limits)
        hw.Hardware.__init__(self, *arks, **kwargs)
        self.label = pc.String(self.name, display=True)

    def set_motor_position(self, motor_position, units="deg"):
        # TODO: should probably support 'motor native units'
        self.q.push("set_motor_position", motor_position)


# --- import --------------------------------------------------------------------------------------


conf = yaqc_cmds.__main__.config
hardwares, gui, advanced_gui = hw.import_hardwares(
    conf.get("hardware", {}).get("filters", {}),
    name="Filters",
    Driver=Driver,
    GUI=GUI,
    Hardware=Hardware,
)



import attune

import somatic.modules.abstract_tuning as abstract_tuning
import yaqc_cmds.project.classes as pc
import yaqc_cmds.hardware.opas as opas


### define ####################################################################


module_name = "TUNE HOLISTIC"


### Worker ####################################################################


class Worker(abstract_tuning.Worker):
    reference_image = "holistic.png"

    def _process(self, data, curve, channel, gtol, ltol, level, scan_folder, config):
        spec = config["Spectral Axis"]["axis"]
        opa = config["OPA"]["opa"]
        motor0 = config["Motor0"]["motor"]
        motor1 = config["Motor1"]["motor"]
        data.transform()
        data.transform(f"{opa}_{motor0}", f"{opa}_{motor1}", spec)
        return attune.workup.holistic(
            data,
            channel,
            [motor0, motor1],
            curve,
            level=level,
            gtol=gtol,
            # ltol=ltol,
            save_directory=scan_folder,
        )


### GUI #######################################################################


class GUI(abstract_tuning.GUI):
    def __init__(self, module_name):
        self.items = {}
        self.items["Spectral Axis"] = abstract_tuning.SpectralAxisSectionWidget(
            "Spectral Axis", self
        )
        self.items["Motor0"] = ProxyMotorAxisSectionWidget("Motor0", self)
        self.items["Motor1"] = abstract_tuning.MotorAxisSectionWidget("Motor1", self)
        super().__init__(module_name)
        self.items["Processing"]["ltol"].set_disabled(True)


class ProxyMotorAxisSectionWidget(abstract_tuning.AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        self.items["Motor"] = pc.Combo()

    def on_update(self):
        hardware = next(h for h in opas.hardwares if h.name == self.parent["OPA"]["OPA"].read())
        self.items["Motor"].set_allowed_values(hardware.curve.dependent_names)


def load():
    return True


def mkGUI():
    global gui
    gui = GUI(module_name)



"""
QFileDialog objects can only be run in the main thread.
"""


### imports ###################################################################


import os
import time

from PySide2 import QtCore
from PySide2 import QtWidgets

from yaqc_cmds.project import project_globals as g
from yaqc_cmds.project import classes as pc


### FileDialog object #########################################################


directory_filepath = pc.Mutex()
open_filepath = pc.Mutex()
save_filepath = pc.Mutex()


class FileDialog(QtCore.QObject):
    update_ui = QtCore.Signal()
    queue_emptied = QtCore.Signal()

    def __init__(self, enqueued_object, busy_object):
        QtCore.QObject.__init__(self)
        self.name = "file_dialog"
        self.enqueued = enqueued_object
        self.busy = busy_object

    @QtCore.Slot(str, list)
    def dequeue(self, method, inputs):
        """
        Slot to accept enqueued commands from main thread.

        Method passed as qstring, inputs as list of [args, kwargs].

        Calls own method with arguments from inputs.
        """
        self.update_ui.emit()
        method = str(method)  # method passed as qstring
        args, kwargs = inputs
        if g.debug.read():
            print(self.name, " dequeue:", method, inputs, self.busy.read())
        self.enqueued.pop()
        getattr(self, method)(*args, **kwargs)
        if not self.enqueued.read():
            self.queue_emptied.emit()
            self.check_busy()

    def check_busy(self):
        """
        decides if the hardware is done and handles writing of 'busy' to False
        """
        # must always write busy whether answer is True or False
        if self.enqueued.read():
            time.sleep(0.1)  # don't loop like crazy
            self.busy.write(True)
        else:
            self.busy.write(False)
            self.update_ui.emit()

    def clean(self, out):
        """
        takes the output and returns a string that has the properties I want
        """
        out = str(out)
        out = out.replace("/", os.sep)
        return out

    def getExistingDirectory(self, inputs=[]):
        caption, directory, options = inputs
        options = QtWidgets.QFileDialog.ShowDirsOnly
        out = self.clean(
            QtWidgets.QFileDialog.getExistingDirectory(
                g.main_window.read(), caption, str(directory), options
            )
        )
        directory_filepath.write(out)

    def getOpenFileName(self, inputs=[]):
        caption, directory, options = inputs
        out = self.clean(
            QtWidgets.QFileDialog.getOpenFileName(
                g.main_window.read(), caption, str(directory), options
            )[0]
        )
        open_filepath.write(out)

    def getSaveFileName(self, inputs=[]):
        caption, directory, savefilter, selectedfilter, options = inputs
        out = self.clean(
            QtWidgets.QFileDialog.getSaveFileName(
                g.main_window.read(),
                caption,
                directory,
                savefilter,
                selectedfilter,
                options,
            )[0]
        )
        save_filepath.write(out)


busy = pc.Busy()
enqueued = pc.Enqueued()
file_dialog = FileDialog(enqueued, busy)
q = pc.Q(enqueued, busy, file_dialog)


### thread-safe file dialog methods ###########################################

# the q method only works between different threads
# call directly if the calling object is in the main thread


def dir_dialog(caption, directory, options=None):
    inputs = [caption, directory, options]
    if QtCore.QThread.currentThread() == g.main_thread.read():
        file_dialog.getExistingDirectory(inputs)
    else:
        q.push("getExistingDirectory", inputs)
        while busy.read():
            time.sleep(0.1)
    return directory_filepath.read()


def open_dialog(caption, directory, options):
    inputs = [caption, directory, options]
    if QtCore.QThread.currentThread() == g.main_thread.read():
        file_dialog.getOpenFileName(inputs)
    else:
        q.push("getOpenFileName", inputs)
        while busy.read():
            time.sleep(0.1)
    return open_filepath.read()


def save_dialog(caption, directory, savefilter, selectedfilter, options):
    inputs = [caption, directory, savefilter, selectedfilter, options]
    if QtCore.QThread.currentThread() == g.main_thread.read():
        file_dialog.getSaveFileName(inputs)
    else:
        q.push("getSaveFileName", inputs)
        while busy.read():
            time.sleep(0.1)
    return save_filepath.read()



import sympy


def const_order(**expressions):
    expressions = {sympy.Symbol(k): sympy.sympify(v) for k, v in expressions.items()}
    while expressions:
        cycle = True
        for symb, expr in list(expressions.items()):
            if all(x.is_constant() or x not in expressions for x in expr.atoms()):
                yield str(symb)
                expressions.pop(symb)
                cycle = False
                break
        if cycle:
            raise ValueError("Cycle detected in set of expressions")



### import ####################################################################

import WrightTools as wt

import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import somatic.acquisition as acquisition
import yaqc_cmds.hardware.opas as opas


### define ####################################################################


module_name = "HOME"


### custom classes ############################################################


class MotorGUI:
    def __init__(self, name, home):
        self.name = name
        self.input_table = pw.InputTable()
        self.input_table.add(name, None)
        self.home = pc.Bool(initial_value=home)
        self.input_table.add("Home", self.home)


class OPA_GUI:
    def __init__(self, hardware, layout):
        self.hardware = hardware
        motor_names = self.hardware.motor_names
        self.motors = []
        for name in motor_names:
            motor = MotorGUI(name, False)
            if layout is not None:
                layout.addWidget(motor.input_table)
            self.motors.append(motor)
        self.hide()  # initialize hidden

    def hide(self):
        for motor in self.motors:
            motor.input_table.hide()

    def show(self):
        for motor in self.motors:
            motor.input_table.show()


### Worker ####################################################################


class Worker(acquisition.Worker):
    def process(self, scan_folder):
        pass

    def run(self):
        # get OPA properties
        opa_name = self.aqn.read("home", "opa name")
        opa_names = [h.name for h in opas.hardwares]
        opa_index = opa_names.index(opa_name)
        opa_hardware = opas.hardwares[opa_index]
        opa_friendly_name = opa_hardware.name
        curve = opa_hardware.curve
        motor_names = self.aqn.read("home", "motor names")
        # motor
        for motor_index, motor_name in enumerate(motor_names):
            if self.aqn.read(motor_name, "home"):
                opa_hardware.home_motor([motor_name])
        opa_hardware.wait_until_still()
        if not self.stopped.read():
            self.finished.write(True)  # only if acquisition successfull


### GUI #######################################################################


class GUI(acquisition.GUI):
    def create_frame(self):
        # shared settings
        input_table = pw.InputTable()
        allowed = [hardware.name for hardware in opas.hardwares]
        self.opa_combo = pc.Combo(allowed)
        input_table.add("OPA", self.opa_combo)
        self.opa_combo.updated.connect(self.on_opa_combo_updated)
        self.layout.addWidget(input_table)
        # motor settings
        self.opa_guis = [OPA_GUI(hardware, self.layout) for hardware in opas.hardwares]
        self.opa_guis[0].show()

    def load(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        # shared settings
        self.opa_combo.write(aqn.read("home", "opa name"))
        # motor settings
        opa = self.opa_guis[self.opa_combo.read_index()]
        for motor, motor_name in zip(opa.motors, aqn.read("home", "motor names")):
            motor.home.write(aqn.read(motor_name, "home"))
        # allow devices to read from aqn
        self.device_widget.load(aqn_path)

    def on_opa_combo_updated(self):
        self.show_opa_gui(self.opa_combo.read_index())

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        opa = self.opa_guis[self.opa_combo.read_index()]
        homed_motor_names = []
        for motor in opa.motors:
            if motor.home.read() == True:
                homed_motor_names.append(motor.name)
        homed_motor_names = str(homed_motor_names).replace("'", "")
        aqn.write(
            "info",
            "description",
            "HOME: {} {}".format(self.opa_combo.read(), homed_motor_names),
        )
        # shared settings
        aqn.add_section("home")
        aqn.write("home", "opa name", self.opa_combo.read())
        aqn.write(
            "home",
            "motor names",
            [motor.name for motor in self.opa_guis[self.opa_combo.read_index()].motors],
        )
        # motor settings
        for motor in opa.motors:
            aqn.add_section(motor.name)
            aqn.write(motor.name, "home", motor.home.read())
        # allow devices to save to aqn
        self.device_widget.save(aqn_path)

    def show_opa_gui(self, index):
        for gui in self.opa_guis:
            gui.hide()
        self.opa_guis[index].show()

    def update_mono_settings(self):
        pass


def mkGUI():
    global gui
    gui = GUI(module_name)


def load():
    return True



# import#########################################################################

import os
import time
import logging
import inspect
import threading
import pathlib
import appdirs

from PySide2 import QtCore

from yaqc_cmds.project import project_globals as g

app = g.app.read()

# cpu watcher####################################################################


class busy(QtCore.QMutex):
    def __init__(self):
        QtCore.QMutex.__init__(self)
        self.WaitCondition = QtCore.QWaitCondition()
        self.value = False

    def read(self):
        return self.value

    def write(self, value):
        self.lock()
        self.value = value
        self.WaitCondition.wakeAll()
        self.unlock()

    def wait_for_update(self):
        if self.value:
            self.lock()
            self.WaitCondition.wait(self)
            self.unlock()


busy = busy()


# logger#########################################################################

# filepath
filepath = (
    pathlib.Path(appdirs.user_log_dir("yaqc-cmds", "yaqc-cmds"))
    / f"{str(time.strftime('%Y%m%dT%H%M%S%z'))}.log"
)
filepath.parent.mkdir(parents=True, exist_ok=True)
log_file = open(filepath, "w")
log_file.close()

# setup
logger = logging.getLogger("Yaqc_cmds")
formatter = logging.Formatter(
    fmt="%(asctime)s.%(msecs).03d | %(levelname)-8s | Thread %(thread)-5s | %(origin)-20s | %(name)-20s | %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S%z",
)

logger.setLevel(logging.DEBUG)


class ContextFilter(logging.Filter):
    def __init__(self, origin, name):
        self.origin = origin
        self.name = name
        super().__init__()

    def filter(self, record):
        record.thread = threading.current_thread().ident
        record.origin = self.origin
        record.name = self.name
        return True


def log(level, name, message="", origin="name"):
    """
    logging method for Yaqc_cmds
    accepts strings
    levels: debug, info, warning, error, critical
    """
    # open
    handler = logging.FileHandler(filepath)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    # additional context
    if origin == "name":
        origin = str(inspect.stack()[2][1]).split(os.sep)[-1].replace(".py", "")
    logger.addFilter(ContextFilter(origin, name))
    # log
    getattr(logger, level)(message)
    # close
    handler.flush()
    logger.removeHandler(handler)
    handler.close()



import os
import pathlib
import ast

import appdirs
import numpy as np
import toml
from PySide2 import QtWidgets, QtCore
import WrightTools as wt
import attune

import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.project_globals as g
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.hardware.delays as delays
import yaqc_cmds.hardware.filters as filters
import yaqc_cmds.somatic as somatic


all_hardwares = {}
all_hardwares.update({h.name: h for h in opas.hardwares})
all_hardwares.update({h.name: h for h in spectrometers.hardwares})
all_hardwares.update({h.name: h for h in delays.hardwares})


class CoSetHW:
    def __init__(self, hardware):
        """
        This object contains all of the GUI and state for one low-level tab of the
        autonomic system.
        """
        self.hardware = hardware
        self.state_path = (
            pathlib.Path(appdirs.user_data_dir("yaqc-cmds", "yaqc-cmds"))
            / "autonomic"
            / f"{self.hardware.name}.toml"
        )
        self.state_path.parent.mkdir(exist_ok=True)
        # instrument
        try:
            self.instrument = attune.load(f"autonomic_{self.hardware.name}")
        except ValueError:
            self.instrument = attune.Instrument(
                arrangements={}, setables={}, name=f"autonomic_{self.hardware.name}"
            )
            attune.store(self.instrument)
        # make own widget
        self.widget = QtWidgets.QWidget()
        self.box = QtWidgets.QHBoxLayout()
        self.box.setContentsMargins(0, 10, 0, 0)
        self.widget.setLayout(self.box)
        self.create_frame(self.box)
        # initialize
        self.visit_store()
        self.update_display()

    def create_frame(self, layout):
        # container widget
        display_container_widget = QtWidgets.QWidget()
        display_container_widget.setLayout(QtWidgets.QVBoxLayout())
        display_layout = display_container_widget.layout()
        display_layout.setMargin(0)
        layout.addWidget(display_container_widget)
        # plot
        self.plot_widget = pw.Plot1D()
        self.plot_scatter = self.plot_widget.add_scatter()
        display_layout.addWidget(self.plot_widget)
        # vertical line
        line = pw.line("V")
        layout.addWidget(line)
        # settings area
        settings_container_widget = QtWidgets.QWidget()
        settings_scroll_area = pw.scroll_area(130)
        settings_scroll_area.setWidget(settings_container_widget)
        settings_scroll_area.setMinimumWidth(300)
        settings_scroll_area.setMaximumWidth(300)
        settings_container_widget.setLayout(QtWidgets.QVBoxLayout())
        settings_layout = settings_container_widget.layout()
        settings_layout.setMargin(5)
        layout.addWidget(settings_scroll_area)
        # input table
        input_table = pw.InputTable()
        input_table.add("Display", None)
        self.hardware_combobox = pc.Combo()  # which control hardware is currently displayed
        self.hardware_combobox.updated.connect(self.on_hardware_combobox_updated)
        input_table.add("Hardware", self.hardware_combobox)
        self.arrangement_combobox = pc.Combo()  # which arrangement of the control hardware
        self.arrangement_combobox.updated.connect(self.update_display)
        input_table.add("Arrangement", self.arrangement_combobox)
        input_table.add("Settings", None)
        input_table.add("Current Offset", self.hardware.offset)
        settings_layout.addWidget(input_table)
        # control arrangement choice table
        self.control_arrangement_table = pw.InputTable()
        self.control_arrangement_bools = {}
        settings_layout.addWidget(self.control_arrangement_table)
        # button
        self.visit_store_button = pw.SetButton("VISIT STORE", color="go")
        g.queue_control.disable_when_true(self.visit_store_button)
        settings_layout.addWidget(self.visit_store_button)
        # stretch
        settings_layout.addStretch(1)

    def launch(self):
        """
        Apply offsets.
        """
        new = 0
        for control in self.instrument.arrangements.keys():
            if not self.control_arrangement_bools[control].read():
                continue
            position = all_hardwares[control].get_position(all_hardwares[control].native_units)
            note = self.instrument(position, arrangement_name=control)
            try:
                key = all_hardwares[control].driver.client.get_arrangement()
                new += note[key]
            except Exception as e:  # TODO: better exception handling
                new += note["auto"]  # default key
        if g.hardware_initialized.read():
            self.hardware.set_offset(new, self.hardware.native_units)

    def on_control_arrangement_bools_updated(self):
        out = {k: v.read() for k, v in self.control_arrangement_bools.items()}
        with open(self.state_path, "w") as f:
            toml.dump(out, f)

    def on_hardware_combobox_updated(self):
        try:
            current_control_arrangement = self.instrument.arrangements[
                self.hardware_combobox.read()
            ]
            self.arrangement_combobox.set_allowed_values(current_control_arrangement.tunes.keys())
        except KeyError:  # probably an instrument without any arrangements
            pass
        self.update_display()

    def update_display(self):
        try:
            arrangement = self.instrument.arrangements[self.hardware_combobox.read()]
            tune = arrangement.tunes[self.arrangement_combobox.read()]
        except KeyError:  # probably an empty instrument
            return
        self.plot_widget.set_labels(tune.ind_units, tune.dep_units)
        self.plot_scatter.clear()
        self.plot_scatter.setData(tune.independent, tune.dependent)

    def visit_store(self):
        self.instrument = attune.load(f"autonomic_{self.hardware.name}")
        # update hardware combobox options
        self.hardware_combobox.set_allowed_values(self.instrument.arrangements.keys())
        self.on_hardware_combobox_updated()
        # add new checkboxes if needed
        for control_name in self.instrument.arrangements.keys():
            if control_name in self.control_arrangement_bools.keys():
                continue
            checkbox = pc.Bool()
            self.control_arrangement_table.add(control_name, checkbox)
            self.control_arrangement_bools[control_name] = checkbox
            checkbox.updated.connect(self.on_control_arrangement_bools_updated)
        # update checkboxes based on saved state
        if not self.state_path.exists():
            self.on_control_arrangement_bools_updated()  # initializes state as false
        for k, v in toml.load(self.state_path).items():
            if k in self.control_arrangement_bools.keys():
                self.control_arrangement_bools[k].write(v)
        self.update_display()

    def zero(self):
        """
        Offsets to zero for all corrs based on current positions.
        """
        instr = self.instrument
        for control in self.instrument.arrangements.keys():
            if not self.control_arrangement_bools[control].read():
                continue
            try:
                tune = all_hardwares[control].driver.client.get_arrangement()
                print(tune)
            except (AttributeError, KeyError) as e:
                print(e)
                tune = "auto"
            print(instr[control].keys(), tune)
            instr = attune.offset_to(
                instr, control, tune, 0, all_hardwares[control].get_position()
            )
        attune.store(instr)
        somatic.signals.updated_attune_store.emit()


coset_hardwares = []  # gets filled by GUI.create_hardware_frame


class Control:
    def __init__(self):
        pass

    def launch(self):
        for coset_hardware in coset_hardwares:
            coset_hardware.launch()

    def visit_store(self):
        for coset_hardware in coset_hardwares:
            coset_hardware.visit_store()

    def zero(self, hardware_name):
        """
        Offsets to zero forr all corrs based on current positions.
        """
        for coset_harware in coset_hardwares:
            if coset_harware.hardware.name == hardware_name:
                coset_harware.zero()
                break


control = Control()
g.hardware_waits.give_coset_control(control)
g.coset_control.write(control)
somatic.signals.updated_attune_store.connect(control.visit_store)


class GUI(QtCore.QObject):
    def __init__(self):
        """Top-level GUI"""
        QtCore.QObject.__init__(self)
        self.create_frame()

    def create_frame(self):
        # get parent layout
        parent_widget = g.coset_widget.read()
        parent_widget.setLayout(QtWidgets.QHBoxLayout())
        parent_widget.layout().setContentsMargins(0, 10, 0, 0)
        parent_layout = parent_widget.layout()
        # create own layout
        self.tabs = QtWidgets.QTabWidget()
        # OPAs
        if len(opas.hardwares) > 0:
            self.create_hardware_frame("OPAs", opas.hardwares)
        # spectrometers
        if len(spectrometers.hardwares) > 0:
            self.create_hardware_frame("Spectrometers", spectrometers.hardwares)
        # delays
        if len(delays.hardwares) > 0:
            self.create_hardware_frame("Delays", delays.hardwares)
        # filters
        if len(filters.hardwares) > 0:
            self.create_hardware_frame("Filters", filters.hardwares)
        parent_layout.addWidget(self.tabs)

    def create_hardware_frame(self, name, hardwares):
        container_widget = QtWidgets.QWidget()
        container_box = QtWidgets.QHBoxLayout()
        container_box.setContentsMargins(0, 10, 0, 0)
        container_widget.setLayout(container_box)
        self.tabs.addTab(container_widget, name)
        # sub-tabs
        tabs = pw.TabWidget()
        container_box.addWidget(tabs)
        for hardware in hardwares:
            coset_hardware = CoSetHW(hardware)
            coset_hardwares.append(coset_hardware)
            tabs.addTab(coset_hardware.widget, hardware.name)


gui = GUI()



__all__ = ["colors", "set_style"]

from PySide2 import QtWidgets
from . import project_globals as g

# see http://www.google.com/design/spec/style/color.html#color-color-palette

colors = {
    "background": "#212121",
    "go": "#009688",
    "stop": "#F44336",
    "set": "#009688",
    "advanced": "#FFC107",
    "text_light": "#EEEEEE",
    "text_disabled": "#757575",
    "widget_background": "#EEEEEE",
    "heading_1": "#00BCD4",
    "heading_0": "#FFC107",
}  # least important heading

g.colors_dict.write(colors)


def set_style():

    # Style Sheet----------------------------------------------------------------

    StyleSheet = ""

    # main window
    StyleSheet += "QMainWindow{background:custom_color}".replace(
        "custom_color", colors["background"]
    )

    # push button
    # StyleSheet += 'QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}'.replace('custom_color', colors['go'])

    # progress bar
    StyleSheet += "QProgressBar:horizontal{border: 0px solid gray; border-radius: 0px; background: custom_color; padding: 0px; height: 30px;}".replace(
        "custom_color", colors["background"]
    )
    StyleSheet += "QProgressBar:chunk{background:custom_color}".replace(
        "custom_color", colors["go"]
    )

    # tab widget
    StyleSheet += "QTabWidget::pane{border-top: 2px solid #C2C7CB;}"
    StyleSheet += "QTabWidget::tab-bar{left: 5px;}"
    StyleSheet += "QTabBar::tab{width: 100px; background: clr1; border: 0px; border-bottom-color: black; border-top-left-radius: 4px; border-top-right-radius: 4px; min-width: 8ex; padding: 2px; font: bold 14px; color: clr2}".replace(
        "clr1", colors["background"]
    ).replace(
        "clr2", colors["heading_0"]
    )
    StyleSheet += "QTabBar::tab:selected{border-color: black; border-bottom-color: black; color: clr1}".replace(
        "clr1", colors["heading_1"]
    )

    # scroll bar
    StyleSheet += "QScrollArea::QWidget::QWidget{backround: transparent;}"

    # group box
    StyleSheet += (
        "QGroupBox{border: 2px solid gray; font: bold 14px; margin-top: 0ex; border-radius: 0 px;}"
    )
    StyleSheet += "QGroupBox::title{subcontrol-origin: margin; padding: 0 0px}"

    app = g.app.read()
    app.setStyleSheet(StyleSheet)

    # Palette--------------------------------------------------------------------

    """
    MainWindow = g.main_window.read()

    palette = QtWidgets.QPalette(MainWindow.palette())
    palette.setColor(MainWindow.backgroundRole(), QtWidgets.QColor(colors['background']))
    #MainWindow.setPalette(palette)

    """

    # style----------------------------------------------------------------------

    # app.setStyle('windowsxp')


def set_background_role(obj):
    palette = QtWidgets.QPalette(obj.palette())
    palette.setColor(obj.backgroundRole(), QtWidgets.QColor(colors["background"]))
    obj.setPalette(palette)



from qtpy import QtCore


class SignalContainer(QtCore.QObject):
    channels_changed = QtCore.Signal()
    sensors_changed = QtCore.Signal()


_signal_container = SignalContainer()
channels_changed = _signal_container.channels_changed
sensors_changed = _signal_container.sensors_changed



"""
Acquisition infrastructure shared by all modules.
"""


### import ####################################################################


import re
import os
import copy
import shutil
import pathlib
import time
import traceback

import appdirs
import toml
import numpy as np

import numexpr

from PySide2 import QtCore, QtWidgets

import WrightTools as wt

import yaqc_cmds.project.project_globals as g

import yaqc_cmds
import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.hardware.delays as delays
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.filters as filters

from yaqc_cmds.somatic._wt5 import create_data, write_data
from yaqc_cmds.somatic.order import ndindex as order
from .signals import data_file_written

all_hardwares = opas.hardwares + spectrometers.hardwares + delays.hardwares + filters.hardwares

from . import constant_resolver


### define ####################################################################


__here__ = pathlib.Path(__file__).parent


### container objects #########################################################


class Axis:
    def __init__(self, points, units, name, hardware_dict={}, **kwargs):
        self.points = points
        self.units = units
        self.name = name
        self.hardware_dict = hardware_dict.copy()
        self.__dict__.update(kwargs)
        # fill hardware dictionary with defaults
        # Not sure if this re.split does what we want now, getting rid of identity
        # KFS 200-10-13
        names = re.split("[=F]+", self.name)
        for name in names:
            if name not in self.hardware_dict.keys():
                hardware_object = next(h for h in all_hardwares if h.name == name)
                self.hardware_dict[name] = [hardware_object, "set_position", None]


class Constant:
    def __init__(self, units, name, static=True, expression=""):
        self.units = units
        self.name = name
        self.static = static
        self.expression = expression
        self.hardware = [h for h in all_hardwares if h.name == self.name][0]


class Destinations:
    def __init__(self, arr, units, hardware, method, passed_args):
        self.arr = arr  # full scan shape
        self.units = units
        self.hardware = hardware
        self.method = method
        self.passed_args = passed_args


### Worker base ##############################################################


class Worker(QtCore.QObject):
    update_ui = QtCore.Signal()
    scan_complete = QtCore.Signal()
    done = QtCore.Signal()

    def __init__(self, aqn_path, queue_worker, finished):
        # do not overload this method
        QtCore.QObject.__init__(self)
        self.aqn_path = aqn_path
        self.aqn = wt.kit.INI(self.aqn_path)
        self.queue_worker = queue_worker
        self.finished = finished
        # unpack
        self.fraction_complete = self.queue_worker.fraction_complete
        self.pause = self.queue_worker.queue_status.pause
        self.paused = self.queue_worker.queue_status.paused
        self.going = self.queue_worker.queue_status.going
        self.stop = self.queue_worker.queue_status.stop
        self.stopped = self.queue_worker.queue_status.stopped
        # move aqn file into queue folder
        ini = wt.kit.INI(aqn_path)
        module_name = ini.read("info", "module")
        item_name = ini.read("info", "name")
        aqn_path = pathlib.Path(aqn_path)
        aqn_index_str = str(self.queue_worker.index.read()).zfill(3)
        aqn_name = " ".join([aqn_index_str, module_name, item_name]).rstrip()
        folder_path = pathlib.Path(self.queue_worker.folder.read()).joinpath(aqn_name)
        if aqn_path != folder_path.with_suffix(".aqn"):
            shutil.copyfile(aqn_path, folder_path.with_suffix(".aqn"))
            if aqn_path.parent == folder_path.parent:
                aqn_path.unlink()
        self.aqn_path = folder_path.with_suffix(".aqn")
        self.aqn = wt.kit.INI(self.aqn_path)
        self.folder = pathlib.Path(folder_path)
        self.folder.mkdir(exist_ok=True)
        # create acquisition folder
        # initialize
        self.scan_index = None
        self.scan_folders = []
        self.scan_urls = []

    def process(self, scan_folder):
        # By default, nothing to do
        return

    def scan(
        self,
        axes,
        constants=[],
        pre_wait_methods=[],
        processing_method="process",
        module_reserved="",
        multiple_scans=False,
    ):
        # do not overload this method
        # scan index ----------------------------------------------------------
        if self.scan_index is None:
            self.scan_index = 0
        else:
            self.scan_index += 1
        # create destination objects ------------------------------------------
        # get destination arrays
        if len(axes) == 1:
            arrs = [axes[0].points]
        else:
            arrs = np.meshgrid(*[a.points for a in axes], indexing="ij")
        # treat 'scan about center' axes
        for axis_index, axis in enumerate(axes):
            if hasattr(axis, "centers"):
                centers = axis.centers
                # transpose so own index is first (all others slide down)
                transpose_order = list(range(len(axes)))
                transpose_order.insert(0, transpose_order.pop(axis_index))
                arrs[axis_index] = np.transpose(arrs[axis_index], axes=transpose_order)
                # add centers to transposed array
                arrs[axis_index] += centers
                # transpose out
                transpose_order = list(range(len(axes)))
                transpose_order.insert(axis_index, transpose_order.pop(0))
                arrs[axis_index] = np.transpose(arrs[axis_index], axes=transpose_order)
        # create destination objects
        destinations_list = []
        for i in range(len(axes)):
            axis = axes[i]
            arr = arrs[i]
            for key in axis.hardware_dict.keys():
                hardware = axis.hardware_dict[key][0]
                method = axis.hardware_dict[key][1]
                passed_args = axis.hardware_dict[key][2]
                destinations = Destinations(arr, axis.units, hardware, method, passed_args)
                destinations_list.append(destinations)
        constant_dict = {c.name: c for c in constants}
        for constant in constant_resolver.const_order(
            **{c.name: c.expression for c in constants}
        ):  # must follow axes
            constant = constant_dict[constant]
            if constant.static:
                pass
            else:
                # initialize
                expression = constant.expression
                arr = np.full(arrs[0].shape, np.nan)
                units = constant.units
                vals = {}
                # populate all hardwares not scanned here
                for hardware in all_hardwares:
                    if wt.units.is_valid_conversion(hardware.units, units):
                        vals[hardware.name] = hardware.get_position(units)
                for idx in np.ndindex(arrs[0].shape):
                    for destination in destinations_list:
                        if wt.units.is_valid_conversion(hardware.units, units):
                            val = wt.units.converter(
                                destination.arr[idx], destination.units, units
                            )
                            vals[destination.hardware.name] = val
                    arr[idx] = numexpr.evaluate(expression, vals)
                # finish
                hardware = constant.hardware
                destinations = Destinations(arr, units, hardware, "set_position", None)
                destinations_list.insert(0, destinations)
        # check if scan is valid for hardware ---------------------------------
        # TODO: !!!
        # run through aquisition order handler --------------------------------
        idxs, slices = order.process(destinations_list)
        # initialize scan -----------------------------------------------------
        g.queue_control.write(True)
        self.going.write(True)
        self.fraction_complete.write(0.0)
        g.logger.log("info", "Scan begun", "")
        # put info into headers -----------------------------------------------
        # create scan folder
        scan_folder = str(self.folder)
        self.scan_folders.append(self.folder)
        # create scan folder on google drive
        if g.google_drive_enabled.read():
            scan_url = g.google_drive_control.read().reserve_id(scan_folder)
            self.scan_urls.append(g.google_drive_control.read().id_to_open_url(scan_folder))
        else:
            self.scan_urls.append(None)
        # create data
        headers = dict()
        headers["name"] = self.aqn.read("info", "name")
        headers["data info"] = self.aqn.read("info", "info")
        headers["data origin"] = self.aqn.read("info", "module")
        if g.google_drive_enabled.read():
            headers["queue url"] = self.queue_worker.queue_url
            headers["acquisition url"] = self.aqn.read("info", "url")
            headers["scan url"] = scan_url
        path = scan_folder + os.sep + "data.wt5"
        create_data(
            path,
            headers,
            destinations,
            axes,
            constants,
            hardware=all_hardwares,
            sensors=yaqc_cmds.sensors.sensors,
        )
        # acquire -------------------------------------------------------------
        self.fraction_complete.write(0.0)
        slice_index = 0
        npts = float(len(idxs))
        for i, idx in enumerate(idxs):
            idx = tuple(idx)
            # launch hardware
            for d in destinations_list:
                destination = d.arr[idx]
                if d.method == "set_position":
                    d.hardware.set_position(destination, d.units)
                else:
                    inputs = copy.copy(d.passed_args)
                    for input_index, input_val in enumerate(inputs):
                        if input_val == "destination":
                            inputs[input_index] = destination
                        elif input_val == "units":
                            inputs[input_index] = d.units
                    d.hardware.q.push(d.method, *inputs)
            # execute pre_wait_methods
            for method in pre_wait_methods:
                method()
            # slice
            if slice_index < len(slices):  # takes care of last slice
                if slices[slice_index]["index"] == i:
                    slice_index += 1
            # wait for hardware
            g.hardware_waits.wait()
            # launch sensors
            for s in yaqc_cmds.sensors.sensors:
                s.measure()
            # wait for sensors
            for s in yaqc_cmds.sensors.sensors:
                s.wait_until_still()
            # save
            write_data(idx=idx, hardware=all_hardwares, sensors=yaqc_cmds.sensors.sensors)
            if i != npts - 1:
                data_file_written.emit()
            # update
            self.fraction_complete.write(i / npts)
            self.update_ui.emit()
            # check continue
            while self.pause.read():
                self.paused.write(True)
                self.pause.wait_for_update()
            self.paused.write(False)
            if self.stop.read():
                self.stopped.write(True)
                break
        # finish scan ---------------------------------------------------------
        self.fraction_complete.write(1.0)
        self.going.write(False)
        g.queue_control.write(False)
        g.logger.log("info", "Scan done", "")
        data_file_written.emit()
        self.update_ui.emit()
        self.scan_complete.emit()
        for s in yaqc_cmds.sensors.sensors:
            s.on_freerun_updated()
        # process scan --------------------------------------------------------
        try:
            getattr(self, processing_method)(scan_folder)
        except BaseException:
            # Yeah, yeah, excepting BaseException.... KFS and BJT
            # deal with it ---sunglasses---  ---BJT 2018-10-25
            traceback.print_exc()
            self.upload(scan_folder)
        return scan_folder

    def upload(self, scan_folder, message="scan complete", reference_image=None):
        # create folder on google drive, upload reference image
        if g.google_drive_enabled.read():
            folder_url = g.google_drive_control.read().id_to_open_url(scan_folder)
            g.google_drive_control.read().upload_folder(
                path=scan_folder,
                parent_id=str(pathlib.Path(scan_folder).parent),
                id=scan_folder,
            )
            image_url = None
            if reference_image is not None:
                reference_id = f"{scan_folder} reference"
                g.google_drive_control.read().reserve_id(reference_id)
                image_url = g.google_drive_control.read().id_to_download_url(reference_id)
                g.google_drive_control.read().create_file(
                    path=reference_image, parent_id=scan_folder, id=reference_id
                )
        else:
            folder_url = image_url = None
        # send message on slack
        if g.slack_enabled.read():
            if g.google_drive_enabled.read() and reference_image is not None:
                start = time.time()
                while time.time() - start < 10 and not g.google_drive_control.read().is_uploaded(
                    reference_id
                ):
                    time.sleep(0.01)
            slack = g.slack_control.read()
            field = {}
            field["title"] = pathlib.Path(scan_folder).name
            field["title_link"] = folder_url
            field["image_url"] = image_url
            message = ":tada: scan complete - {} elapsed".format(
                g.progress_bar.time_elapsed.text()
            )
            slack.send_message(message, attachments=[field])


### GUI base ##################################################################


class GUI(QtCore.QObject):
    """Acquisition module gui."""

    def __init__(self, module_name):
        QtCore.QObject.__init__(self)
        self.module_name = module_name
        self.state_path = (
            pathlib.Path(appdirs.user_data_dir("yaqc-cmds", "yaqc-cmds"))
            / "modules"
            / f"{self.module_name.lower()}.toml"
        )
        self.state_path.parent.mkdir(parents=True, exist_ok=True)
        self.state_path.touch(exist_ok=True)
        self.state = toml.load(self.state_path)
        # create frame
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setMargin(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.create_frame()  # add module-specific widgets to out layout
        # device widget
        # self.device_widget = record.Widget()
        # self.layout.addWidget(self.device_widget)
        # finish
        self.frame = QtWidgets.QWidget()
        self.frame.setLayout(self.layout)

    def create_frame(self):
        layout = QtWidgets.QVBoxLayout()
        layout.setMargin(5)
        # scan widget
        layout.addWidget(self.scan.widget)
        # finish
        layout.addStretch(1)
        self.frame = QtWidgets.QWidget()
        self.frame.setLayout(layout)
        g.module_widget.add_child(self.frame)
        g.module_combobox.add_module(module_name, self.show_frame)

    def hide(self):
        self.frame.hide()

    def on_sensor_settings_updated(self):
        # overload this if your gui has device-dependent settings
        pass

    def save_state(self):
        with open(self.state_path, "w") as f:
            f.write(toml.dumps(self.state))

    def show(self):
        self.frame.show()

    def update(self):
        pass



### import ####################################################################


import numpy as np

import matplotlib

matplotlib.pyplot.ioff()

from PySide2 import QtWidgets
import WrightTools as wt
import attune

import yaqc_cmds.project.project_globals as g
import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.somatic.acquisition as acquisition
from yaqc_cmds.somatic.modules.scan import Axis as ScanAxisGUI
from yaqc_cmds.somatic.modules.scan import Constant

import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.hardware.delays as delays
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.filters.filters as filters

all_hardwares = opas.hardwares + spectrometers.hardwares + delays.hardwares + filters.hardwares
import yaqc_cmds.devices.devices as devices


### define ####################################################################


module_name = "ZERO TUNE"


### Worker ####################################################################


class Worker(acquisition.Worker):
    def process(self, scan_folder):
        data_path = wt.kit.glob_handler(".data", folder=str(scan_folder))[0]
        data = wt.data.from_Yaqc_cmds(data_path)
        delays = self.aqn.read("delay", "delays")
        channel_name = self.aqn.read("processing", "channel")
        transform = list(data.axis_expressions)[:2]
        for axis in data.axis_expressions:
            if axis not in transform:
                if level:
                    data.level(axis, 0, 5)
                data.moment(axis, channel)
                channel_name = -1
        for delay in delays:
            attune.workup.intensity(
                data,
                channel_name,
                delay,
                level=self.aqn.read("processing", "level"),
                gtol=self.aqn.read("processing", "gtol"),
                ltol=self.aqn.read("processing", "ltol"),
                save_directory=scan_folder,
            )

        # upload
        self.upload(scan_folder)

    def run(self):
        axes = []
        # OPA
        opa_name = self.aqn.read("opa", "opa")
        npts = self.aqn.read("opa", "npts")
        opa_names = [opa.name for opa in opas.hardwares]
        opa_index = opa_names.index(opa_name)
        opa_hardware = opas.hardwares[opa_index]
        opa_friendly_name = opa_hardware.name
        curve = opa_hardware.curve.copy()
        curve.convert("wn")
        pts = np.linspace(curve.setpoints[:].min(), curve.setpoints[:].max(), int(npts))
        axis = acquisition.Axis(pts, "wn", opa_friendly_name, opa_friendly_name)
        axes.append(axis)
        # delay
        axis_name = "delay"
        start = self.aqn.read(axis_name, "start")
        stop = self.aqn.read(axis_name, "stop")
        number = self.aqn.read(axis_name, "number")
        points = np.linspace(start, stop, int(number))
        units = self.aqn.read(axis_name, "units")
        name = "=".join(self.aqn.read(axis_name, "delays"))
        axis = acquisition.Axis(points, units, name, name)
        axes.append(axis)

        constants = []
        for constant_name in self.aqn.read("scan", "constant names"):
            for hardware in all_hardwares:
                if hardware.name == constant_name:
                    units = hardware.units
                    if wt.units.kind(units) == "energy":
                        units = "wn"
                    break
            name = constant_name
            identity = expression = self.aqn.read(constant_name, "expression")
            constant = acquisition.Constant(
                units, name, identity, expression=expression, static=False
            )
            constants.append(constant)
        # do scan
        self.scan(axes, constants)
        # finish
        if not self.stopped.read():
            self.finished.write(True)  # only if acquisition successfull


### GUI #######################################################################


class GUI(acquisition.GUI):
    def create_frame(self):
        input_table = pw.InputTable()
        # opa combo
        allowed = [hardware.name for hardware in opas.hardwares]
        self.opa_combo = pc.Combo(allowed)
        input_table.add("OPA", None)
        input_table.add("OPA", self.opa_combo)
        self.npts_opa = pc.Number(decimals=0, initial_value=21)
        input_table.add("npts", self.npts_opa)
        # delay
        self.delay = ScanAxisGUI("delay", "")
        self.delay.start.write(-3)
        self.delay.stop.write(3)
        self.delay.number.write(21)
        input_table.add("Delay", None)
        self.layout.addWidget(input_table)
        self.layout.addWidget(self.delay.widget)
        # constants
        self.constants = []
        input_table = pw.InputTable()
        input_table.add("Constants", None)
        self.layout.addWidget(input_table)
        self.constants_container_widget = QtWidgets.QWidget()
        self.constants_container_widget.setLayout(QtWidgets.QVBoxLayout())
        self.constants_container_widget.layout().setMargin(0)
        self.layout.addWidget(self.constants_container_widget)
        add_constant_button, remove_constant_button = self.add_buttons()
        add_constant_button.clicked.connect(self.add_constant)
        remove_constant_button.clicked.connect(self.remove_constant)
        # processing
        input_table = pw.InputTable()
        input_table.add("Processing", None)
        self.channel_combo = pc.Combo(allowed_values=devices.control.channel_names)
        input_table.add("Channel", self.channel_combo)
        self.process_level = pc.Bool(initial_value=False)
        self.process_gtol = pc.Number(initial_value=0, decimals=5)
        self.process_ltol = pc.Number(initial_value=1e-2, decimals=5)
        input_table.add("level", self.process_level)
        input_table.add("gtol", self.process_gtol)
        input_table.add("ltol", self.process_ltol)
        # finish
        self.layout.addWidget(input_table)

    def add_constant(self):
        # if len(self.constants) == 1: return  # temporary...
        constant = Constant()
        self.constants_container_widget.layout().addWidget(constant.widget)
        self.constants.append(constant)

    def add_buttons(self):
        colors = g.colors_dict.read()
        # layout
        button_container = QtWidgets.QWidget()
        button_container.setLayout(QtWidgets.QHBoxLayout())
        button_container.layout().setMargin(0)
        # remove
        remove_button = QtWidgets.QPushButton()
        remove_button.setText("REMOVE")
        remove_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["stop"]
        )
        remove_button.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(remove_button)
        # add
        add_button = QtWidgets.QPushButton()
        add_button.setText("ADD")
        add_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["set"]
        )
        add_button.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(add_button)
        # finish
        self.layout.addWidget(button_container)
        return [add_button, remove_button]

    def load(self, aqn_path):
        for constant in self.constants:
            constant.hide()
        self.constants = []
        aqn = wt.kit.INI(aqn_path)
        self.opa_combo.write(aqn.read("opa", "opa"))
        self.npts_opa.write(aqn.read("opa", "npts"))
        self.channel_combo.write(aqn.read("processing", "channel"))
        self.process_level.write(aqn.read("process", "level"))
        self.process_gtol.write(aqn.read("process", "gtol"))
        self.process_ltol.write(aqn.read("process", "ltol"))
        self.delay.start.write(aqn.read("delay", "start"))
        self.delay.stop.write(aqn.read("delay", "stop"))
        self.delay.number.write(aqn.read("delay", "number"))
        for key, mutex in self.delay.hardwares.items():
            mutex.write(key in aqn.read("delay", "delays"))
        # constants
        constant_names = aqn.read("scan", "constant names")

        for constant_index, constant_name in enumerate(constant_names):
            constant = Constant()
            constant.hardware_name_combo.write(aqn.read(constant_name, "hardware"))
            constant.expression.write(aqn.read(constant_name, "expression"))
            self.constants.append(constant)
            self.constants_container_widget.layout().addWidget(constant.widget)
        # allow devices to load settings
        self.device_widget.load(aqn_path)

    def remove_constant(self):
        # remove trailing constant
        if len(self.constants) > 0:
            constant = self.constants[-1]
            self.constants_container_widget.layout().removeWidget(constant.widget)
            constant.hide()
            self.constants.pop(-1)

    def on_device_settings_updated(self):
        self.channel_combo.set_allowed_values(devices.control.channel_names)

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        aqn.add_section("opa")
        aqn.write("opa", "opa", self.opa_combo.read())
        aqn.write("opa", "npts", self.npts_opa.read())
        aqn.add_section("delay")
        aqn.write("delay", "start", self.delay.start.read())
        aqn.write("delay", "stop", self.delay.stop.read())
        aqn.write("delay", "number", self.delay.number.read())
        aqn.write("delay", "units", self.delay.units)
        hardwares = []
        for key, bool_mutex in self.delay.hardwares.items():
            if bool_mutex.read():
                hardwares.append(key)
        aqn.write("delay", "delays", hardwares)
        aqn.write(
            "info",
            "description",
            "{} {} zero tune".format(self.opa_combo.read(), hardwares),
        )
        # constants
        aqn.add_section("scan")
        aqn.write("scan", "constant names", [c.get_name() for c in self.constants])
        for constant in self.constants:
            name = constant.get_name()
            aqn.add_section(name)
            aqn.write(name, "hardware", constant.hardware_name_combo.read())
            aqn.write(name, "expression", constant.expression.read())
        aqn.add_section("processing")
        aqn.write("processing", "channel", self.channel_combo.read())
        aqn.write("processing", "level", self.process_level.read())
        aqn.write("processing", "gtol", self.process_gtol.read())
        aqn.write("processing", "ltol", self.process_ltol.read())
        # allow devices to write settings
        self.device_widget.save(aqn_path)


def load():
    return True


def mkGUI():
    global gui
    gui = GUI(module_name)



import time

import yaqc_cmds.project.slack.bots as bots

channel = "C0DD6BG9L"

ctrl = bots.witch


imageid = "0B8z-JGr_8g4RSlNPMC04bVBSZUk"
url = "https://drive.google.com/open?id=0B8z-JGr_8g4RdTN4SHdxVGZTSkE"
image_url = "https://docs.google.com/uc?id=" + imageid

field0 = {}
field0["title"] = "MOTORTUNE [w1, w1_Mixer_2, wm] 2016.02.02 16_03_57"
field0["title_link"] = url
field0["image_url"] = image_url

ctrl.send_message("scan complete - 00:09:51 elapsed", channel, attachments=[field0])



### import ####################################################################


import yaqc_cmds.__main__
import yaqc_cmds.project.classes as pc
import yaqc_cmds.hardware.hardware as hw
import pathlib
import appdirs
import toml
import yaqc


### driver ####################################################################


class Driver(hw.Driver):
    def __init__(self, *args, **kwargs):
        self._yaqd_port = kwargs.pop("yaqd_port")
        super().__init__(*args, **kwargs)

        # open control
        self.ctrl = yaqc.Client(self._yaqd_port)
        # import some information from control
        id_dict = self.ctrl.id()
        self.serial_number = id_dict["serial"]
        self.position.write(self.ctrl.get_position())
        # recorded
        self.recorded[self.name] = [self.position, self.native_units, 1.0, "m", False]
        self.limits.write(*self.ctrl.get_limits())
        self.wait_until_still()
        # finish
        self.initialized.write(True)
        self.initialized_signal.emit()

        self.grating = pc.Combo(
            name="Grating",
            allowed_values=self.ctrl.get_turret_options(),
            section=self.name,
            option="grating",
            display=True,
            set_method="set_turret",
        )
        self.exposed.append(self.grating)

    def get_position(self):
        native_position = self.ctrl.get_position()
        self.position.write(native_position, self.native_units)
        return self.position.read()

    def is_busy(self):
        return self.ctrl.busy()

    def set_position(self, destination):
        self.ctrl.set_position(float(destination))
        self.wait_until_still()

    def set_turret(self, destination):
        if type(destination) == list:
            destination = destination[0]
        self.ctrl.set_turret(destination)
        self.grating.write(destination)
        self.wait_until_still()
        self.limits.write(*self.ctrl.get_limits(), self.native_units)


### gui #######################################################################


class GUI(hw.GUI):
    pass


### hardware ##################################################################


class Hardware(hw.Hardware):
    def __init__(self, *args, **kwargs):
        self.kind = "spectrometer"
        hw.Hardware.__init__(self, *args, **kwargs)


### import ####################################################################


conf = yaqc_cmds.__main__.config
hardwares, gui, advanced_gui = hw.import_hardwares(
    conf.get("hardware", {}).get("spectrometers", {}),
    name="Spectrometers",
    Driver=Driver,
    GUI=GUI,
    Hardware=Hardware,
)



import pathlib

import appdirs
import toml

from . import signals
from ._sensors import Sensor, Driver, SensorWidget
import yaqc_cmds.__main__


sensors = []
config = yaqc_cmds.__main__.config

for section in config["sensors"].keys():
    if section == "settings":
        continue
    if config["sensors"][section]["enable"]:
        # collect arguments
        kwargs = dict()
        for option in config["sensors"][section].keys():
            if option in ["enable", "model", "serial", "path", "__name__"]:
                continue
            else:
                kwargs[option] = config["sensors"][section][option]
        sensor = Sensor(
            Driver,
            kwargs,
            None,
            Widget=SensorWidget,
            name=section,
            model="yaq",
        )
        sensors.append(sensor)
        sensor.initialize()


def get_channels_dict():
    channels = dict()
    for sensor in sensors:
        for name in sensor.channel_names:
            channels[name] = sensor
    if not channels:
        raise Exception("you must have at least one sensor with channels to run yaqc-cmds")
    return channels



### import ####################################################################


import os
import time
import pathlib

import numpy as np

from PySide2 import QtCore

from . import project_globals as g

import WrightTools.units as wt_units


__here__ = pathlib.Path(__file__).parent


### mutex #####################################################################


class Mutex(QtCore.QMutex):
    def __init__(self, initial_value=None):
        QtCore.QMutex.__init__(self)
        self.WaitCondition = QtCore.QWaitCondition()
        self.value = initial_value

    def read(self):
        return self.value

    def write(self, value):
        self.lock()
        self.value = value
        self.WaitCondition.wakeAll()
        self.unlock()

    def wait_for_update(self, timeout=5000):
        if self.value:
            self.lock()
            self.WaitCondition.wait(self, timeout)
            self.unlock()


class Busy(QtCore.QMutex):
    def __init__(self):
        """
        QMutex object to communicate between threads that need to wait \n
        while busy.read(): busy.wait_for_update()
        """
        QtCore.QMutex.__init__(self)
        self.WaitCondition = QtCore.QWaitCondition()
        self.value = False
        self.type = "busy"
        self.update_signal = None

    def read(self):
        return self.value

    def write(self, value):
        """
        bool value
        """
        self.lock()
        self.value = value
        self.WaitCondition.wakeAll()
        self.unlock()

    def wait_for_update(self, timeout=5000):
        """
        wait in calling thread for any thread to call 'write' method \n
        int timeout in milliseconds
        """
        if self.value:
            self.lock()
            self.WaitCondition.wait(self, timeout)
            self.unlock()


### gui items #################################################################


class Value(QtCore.QMutex):
    def __init__(self, initial_value=None):
        """
        Basic QMutex object to hold a single object in a thread-safe way.
        """
        QtCore.QMutex.__init__(self)
        self.value = initial_value

    def read(self):
        return self.value

    def write(self, value):
        self.lock()
        self.value = value
        self.unlock()


class PyCMDS_Object(QtCore.QObject):
    updated = QtCore.Signal()
    disabled = False

    def __init__(
        self,
        initial_value=None,
        display=False,
        name="",
        label="",
        set_method=None,
        disable_under_queue_control=False,
        *args,
        **kwargs,
    ):
        QtCore.QObject.__init__(self)
        self.has_widget = False
        self.tool_tip = ""
        self.value = Value(initial_value)
        self.display = display
        self.set_method = set_method
        if self.display:
            self.disabled = True
        else:
            self.disabled = False
        # name
        self.name = name
        if not label == "":
            pass
        else:
            self.label = self.name
        # disable under module control
        if disable_under_queue_control:
            g.main_window.read().queue_control.connect(self.on_queue_control)

    def associate(self, display=None, pre_name=""):
        # display
        if display is None:
            display = self.display
        # name
        name = pre_name + self.name
        # new object
        new_obj = self.__class__(initial_value=self.read(), display=display, name=name)
        return new_obj

    def on_queue_control(self):
        if g.queue_control.read():
            if self.has_widget:
                self.widget.setDisabled(True)
        else:
            if self.has_widget:
                self.widget.setDisabled(self.disabled)

    def read(self):
        return self.value.read()

    def write(self, value):
        self.value.write(value)
        self.updated.emit()

    def set_disabled(self, disabled):
        self.disabled = bool(disabled)
        if self.has_widget:
            self.widget.setDisabled(self.disabled)

    def setDisabled(self, disabled):
        self.set_disabled(disabled)

    def set_tool_tip(self, tool_tip):
        self.tool_tip = str(tool_tip)
        if self.has_widget:
            self.widget.setToolTip(self.tool_tip)


class Bool(PyCMDS_Object):
    """
    holds 'value' (bool) - the state of the checkbox

    use read method to access
    """

    def __init__(self, initial_value=False, *args, **kwargs):
        PyCMDS_Object.__init__(self, initial_value=initial_value, *args, **kwargs)
        self.type = "checkbox"

    def give_control(self, control_widget):
        self.widget = control_widget
        # set
        self.widget.setChecked(self.value.read())
        # connect signals and slots
        self.updated.connect(lambda: self.widget.setChecked(self.value.read()))
        self.widget.stateChanged.connect(lambda: self.write(self.widget.isChecked()))
        # finish
        self.widget.setToolTip(self.tool_tip)
        self.widget.setDisabled(self.disabled)
        self.has_widget = True


class Combo(PyCMDS_Object):
    def __init__(self, allowed_values=["None"], initial_value=None, *args, **kwargs):
        PyCMDS_Object.__init__(self, *args, **kwargs)
        self.type = "combo"
        self.allowed_values = list(allowed_values)
        self.data_type = type(self.allowed_values[0])
        if initial_value is None:
            self.write(self.allowed_values[0])
        else:
            self.write(initial_value)

    def associate(self, display=None, pre_name=""):
        # display
        if display is None:
            display = self.display
        # name
        name = pre_name + self.name
        # new object
        new_obj = Combo(
            initial_value=self.read(),
            display=display,
            allowed_values=self.allowed_values,
            name=name,
        )
        return new_obj

    def read_index(self):
        return self.allowed_values.index(self.read())

    def set_allowed_values(self, allowed_values):
        """
        Set the allowed values of the Combo object.

        Parameters
        ----------
        allowed_values : list
            the new allowed values

        Notes
        ----------
        The value of the object is written to the first allowed value if the
        current value is not in the allowed values.
        """
        if allowed_values == self.allowed_values:
            return
        if not allowed_values:  # case of empty list
            return
        self.allowed_values = list(allowed_values)
        # update widget
        if self.has_widget:
            self.widget.currentIndexChanged.disconnect(self.write_from_widget)
            self.widget.clear()
            allowed_values_strings = [str(value) for value in self.allowed_values]
            self.widget.addItems(allowed_values_strings)
            self.widget.currentIndexChanged.connect(self.write_from_widget)
        # write value again
        if self.read() not in self.allowed_values:
            self.write(list(self.allowed_values)[0])
        else:
            self.write(self.read())

    def set_widget(self):
        allowed_values_strings = [str(value) for value in self.allowed_values]
        index = allowed_values_strings.index(str(self.read()))
        self.widget.setCurrentIndex(index)

    def write(self, value):
        # value will be maintained as original data type
        value = self.data_type(value)
        PyCMDS_Object.write(self, value)

    def write_from_widget(self):
        # needs to be defined method so we can connect and disconnect
        self.write(self.widget.currentText())

    def give_control(self, control_widget):
        self.widget = control_widget
        # fill out items
        allowed_values_strings = [str(value) for value in self.allowed_values]
        self.widget.addItems(allowed_values_strings)
        if self.read() is not None:
            self.widget.setCurrentIndex(allowed_values_strings.index(str(self.read())))
        # connect signals and slots
        self.updated.connect(self.set_widget)
        self.widget.currentIndexChanged.connect(self.write_from_widget)
        self.widget.setToolTip(self.tool_tip)
        self.widget.setDisabled(self.disabled)
        self.has_widget = True


class Filepath(PyCMDS_Object):
    def __init__(self, caption="Open", directory=None, options=[], kind="file", *args, **kwargs):
        """
        holds the filepath as a string \n

        Kind one in {'file', 'directory'}
        """
        PyCMDS_Object.__init__(self, *args, **kwargs)
        self.type = "filepath"
        self.caption = caption
        self.directory = directory
        self.options = options
        self.kind = kind

    def give_control(self, control_widget):
        self.widget = control_widget
        if self.read() is not None:
            self.widget.setText(self.read())
        # connect signals and slots
        self.updated.connect(lambda: self.widget.setText(self.read()))
        self.widget.setToolTip(str(self.read()))
        self.updated.connect(lambda: self.widget.setToolTip(self.read()))
        self.has_widget = True

    def give_button(self, button_widget):
        self.button = button_widget
        self.button.clicked.connect(self.on_load)

    def on_load(self):
        from project import file_dialog_handler

        # directory
        if self.directory is not None:
            directory_string = self.directory
        else:
            if self.read() is not None:
                directory_string = self.read()
            else:
                directory_string = __here__.parent
        # filter

        if self.kind == "file":
            filter_string = ";;".join(self.options + ["All Files (*.*)"])
            out = file_dialog_handler.open_dialog(self.caption, directory_string, filter_string)
            if os.path.isfile(out):
                self.write(out)
        elif self.kind == "directory":
            out = file_dialog_handler.dir_dialog(self.caption, directory_string, "")
            if os.path.isdir(out):
                self.write(out)

    def read(self):
        # want python string, not QString
        return str(PyCMDS_Object.read(self))


class NumberLimits(PyCMDS_Object):
    def __init__(self, min_value=-1e6, max_value=1e6, units=None):
        """
        not appropriate for use as a gui element - only for backend use
        units must never change for this kind of object
        """
        PyCMDS_Object.__init__(self)
        PyCMDS_Object.write(self, [min_value, max_value])
        self.units = units

    def read(self, output_units="same"):
        min_value, max_value = PyCMDS_Object.read(self)
        if output_units == "same":
            pass
        else:
            min_value = wt_units.converter(min_value, self.units, output_units)
            max_value = wt_units.converter(max_value, self.units, output_units)
        # ensure order
        min_value, max_value = [
            min([min_value, max_value]),
            max([min_value, max_value]),
        ]
        return [min_value, max_value]

    def write(self, min_value, max_value, input_units="same"):
        if input_units == "same":
            pass
        else:
            min_value = wt_units.converter(min_value, input_units, self.units)
            max_value = wt_units.converter(max_value, input_units, self.units)
        # ensure order
        min_value, max_value = [
            min([min_value, max_value]),
            max([min_value, max_value]),
        ]
        PyCMDS_Object.write(self, [min_value, max_value])
        self.updated.emit()


class Number(PyCMDS_Object):
    units_updated = QtCore.Signal()

    def __init__(
        self,
        initial_value=np.nan,
        single_step=1.0,
        decimals=3,
        limits=None,
        units=None,
        *args,
        **kwargs,
    ):
        PyCMDS_Object.__init__(self, initial_value=initial_value, *args, **kwargs)
        self.type = "number"
        self.disabled_units = False
        self.single_step = single_step
        self.decimals = decimals
        self.set_control_steps(single_step, decimals)
        # units
        self.units = units
        # limits
        self.limits = limits
        if self.limits is None:
            self.limits = NumberLimits()
        if self.units is None:
            self.limits.units = None
        if self.units is not None and self.limits.units is None:
            self.limits.units = self.units
        self._set_limits()
        self.limits.updated.connect(self._set_limits)

    def _set_limits(self):
        min_value, max_value = self.limits.read()
        limits_units = self.limits.units
        min_value = wt_units.converter(min_value, limits_units, self.units)
        max_value = wt_units.converter(max_value, limits_units, self.units)
        # ensure order
        min_value, max_value = [
            min([min_value, max_value]),
            max([min_value, max_value]),
        ]
        if self.has_widget:
            self.widget.setMinimum(min_value)
            self.widget.setMaximum(max_value)
            if not self.display:
                self.set_tool_tip("min: " + str(min_value) + "\n" + "max: " + str(max_value))

    def associate(self, display=None, pre_name=""):
        # display
        if display is None:
            display = self.display
        # name
        name = pre_name + self.name
        # new object
        new_obj = Number(
            initial_value=self.read(),
            display=display,
            units=self.units,
            limits=self.limits,
            single_step=self.single_step,
            decimals=self.decimals,
            name=name,
        )
        return new_obj

    def convert(self, destination_units):
        # value
        self.value.lock()
        old_val = self.value.read()
        new_val = wt_units.converter(old_val, self.units, str(destination_units))
        self.value.unlock()
        self.value.write(new_val)
        # commit and signal
        self.units = str(destination_units)
        self._set_limits()
        self.units_updated.emit()
        self.updated.emit()

    def read(self, output_units="same"):
        value = PyCMDS_Object.read(self)
        if output_units == "same":
            pass
        else:
            value = wt_units.converter(value, self.units, output_units)
        return value

    def set_control_steps(self, single_step=None, decimals=None):
        limits = [self.single_step, self.decimals]
        inputs = [single_step, decimals]
        widget_methods = ["setSingleStep", "setDecimals"]
        for i in range(len(limits)):
            if not inputs[i] is None:
                limits[i] = inputs[i]
            if self.has_widget:
                getattr(self.widget, widget_methods[i])(limits[i])

    def set_disabled_units(self, disabled):
        self.disabled_units = bool(disabled)
        if self.has_widget:
            self.units_widget.setDisabled(self.disabled_units)

    def set_units(self, units):
        if self.has_widget:
            allowed = [self.units_widget.itemText(i) for i in range(self.units_widget.count())]
            index = allowed.index(units)
            self.units_widget.setCurrentIndex(index)
        else:
            self.convert(units)

    def set_widget(self):
        # special value text is displayed when widget is at minimum
        if np.isnan(self.value.read()):
            self.widget.setSpecialValueText("nan")
            self.widget.setValue(self.widget.minimum())
        else:
            self.widget.setSpecialValueText("")
            self.widget.setValue(self.value.read())

    def give_control(self, control_widget):
        self.widget = control_widget
        # set values
        min_value, max_value = self.limits.read()
        self.widget.setMinimum(min_value)
        self.widget.setMaximum(max_value)
        self.widget.setDecimals(self.decimals)
        self.widget.setSingleStep(self.single_step)
        self.set_widget()
        # connect signals and slots
        self.updated.connect(self.set_widget)
        self.widget.editingFinished.connect(lambda: self.write(self.widget.value()))
        # finish
        self.widget.setToolTip(self.tool_tip)
        self.widget.setDisabled(self.disabled)
        self.has_widget = True
        self._set_limits()

    def give_units_combo(self, units_combo_widget):
        self.units_widget = units_combo_widget
        # add items
        unit_types = [self.units] + list(wt_units.get_valid_conversions(self.units))
        self.units_widget.addItems(unit_types)
        # set current item
        self.units_widget.setCurrentIndex(unit_types.index(self.units))
        # associate update with conversion
        self.units_widget.currentIndexChanged.connect(
            lambda: self.convert(self.units_widget.currentText())
        )
        # finish
        self.units_widget.setDisabled(self.disabled_units)

    def write(self, value, input_units="same"):
        if input_units == "same":
            pass
        else:
            value = wt_units.converter(value, input_units, self.units)
        PyCMDS_Object.write(self, value)


class String(PyCMDS_Object):
    def __init__(self, initial_value="", max_length=None, *args, **kwargs):
        PyCMDS_Object.__init__(self, initial_value=initial_value, *args, **kwargs)
        self.type = "string"
        self.max_length = max_length

    def give_control(self, control_widget):
        self.widget = control_widget
        if self.max_length is not None:
            self.widget.setMaxLength(self.max_length)
        # fill out items
        self.widget.setText(str(self.value.read()))
        # connect signals and slots
        self.updated.connect(lambda: self.widget.setText(self.value.read()))
        self.widget.editingFinished.connect(lambda: self.write(str(self.widget.text())))
        self.widget.setToolTip(self.tool_tip)
        self.has_widget = True

    def read(self):
        return str(PyCMDS_Object.read(self))

    def write(self, value):
        if self.max_length is not None:
            value = value[: self.max_length]
        self.value.write(value)
        self.updated.emit()


### part ######################################################################


class Driver(QtCore.QObject):
    update_ui = QtCore.Signal()
    queue_emptied = QtCore.Signal()
    initialized = Bool()

    def check_busy(self):
        """
        Handles writing of busy to False.

        Must always write to busy.
        """
        if self.is_busy():
            if g.debug.read():
                print(self.name, " busy")
            time.sleep(0.01)  # don't loop like crazy
            self.busy.write(True)
        elif self.enqueued.read():
            if g.debug.read():
                print(self.name, " not empty")
                print(self.enqueued.value)
            time.sleep(0.1)  # don't loop like crazy
            self.busy.write(True)
        else:
            if g.debug.read():
                print(self.name, " not busy")
            self.busy.write(False)
            self.update_ui.emit()

    @QtCore.Slot(str, list)
    def dequeue(self, method, inputs):
        """
        Slot to accept enqueued commands from main thread.

        Method passed as qstring, inputs as list of [args, kwargs].

        Calls own method with arguments from inputs.
        """
        self.update_ui.emit()
        method = str(method)  # method passed as qstring
        args, kwargs = inputs
        if g.debug.read():
            print(self.name, " dequeue:", method, inputs, self.busy.read())
        self.enqueued.pop()
        getattr(self, method)(*args, **kwargs)
        if not self.enqueued.read():
            if g.debug.read():
                print(self.name, " queue empty")
            self.queue_emptied.emit()
            self.check_busy()

    def is_busy(self):
        return False


class Enqueued(QtCore.QMutex):
    def __init__(self):
        """
        Holds list of enqueued options.
        """
        QtCore.QMutex.__init__(self)
        self.value = []

    def read(self):
        return self.value

    def push(self, value):
        self.lock()
        self.value.append(value)
        self.unlock()

    def pop(self):
        self.lock()
        self.value = self.value[1:]
        self.unlock()


class Hardware(QtCore.QObject):
    update_ui = QtCore.Signal()
    initialized_signal = QtCore.Signal()

    def __init__(self, driver_class, driver_arguments, gui_class, name, model, serial=None):
        """
        Hardware representation object living in the main thread.

        Parameters
        driver_class : Driver class
            Class of driver.
        driver_arguments : dictionary
            Arguments passed to driver upon initialization.
        name : string
            Name. Must be unique.
        model : string
            Model. Need not be unique.
        serial : string or None (optional)
            Serial, if desired. Default is None.
        """
        QtCore.QObject.__init__(self)
        self.name = name
        self.model = model
        self.serial = serial
        # create objects
        self.thread = QtCore.QThread()
        self.enqueued = Enqueued()
        self.busy = Busy()
        self.driver = driver_class(self, **driver_arguments)
        self.initialized = self.driver.initialized
        if gui_class:
            self.gui = gui_class(self)
        self.q = Q(self.enqueued, self.busy, self.driver)
        # start thread
        self.driver.moveToThread(self.thread)
        self.thread.start()
        # connect to address object signals
        self.driver.update_ui.connect(self.update)
        self.busy.update_signal = self.driver.update_ui
        # initialize drivers
        self.q.push("initialize")
        # integrate close into PyCMDS shutdown
        self.shutdown_timeout = 30  # seconds
        g.shutdown.add_method(self.close)
        g.hardware_waits.add(self.wait_until_still)

    def close(self):
        # begin driver shutdown
        self.q.push("close")
        # wait for driver shutdown to complete
        start_time = time.time()
        self.q.push("check_busy")
        while self.busy.read():
            if time.time() - start_time < self.shutdown_timeout:
                self.busy.wait_for_update()
            else:
                g.logger.log("warning", "Wait until done timed out", self.name)
                break
        # quit thread
        self.thread.exit()
        self.thread.quit()

    def update(self):
        self.update_ui.emit()

    def wait_until_still(self):
        while self.busy.read():
            if not self.q.enqueued.value:
                self.q.push("check_busy")
            self.busy.wait_for_update()


class Q(QtCore.QObject):
    signal = QtCore.Signal(str, list)

    def __init__(self, enqueued, busy, driver):
        self.enqueued = enqueued
        self.busy = busy
        self.driver = driver
        super(Q, self).__init__()
        # self.queue = QtCore.QMetaObject()
        self.signal.connect(self.driver.dequeue, type=QtCore.Qt.QueuedConnection)

    def push(self, method, *args, **kwargs):
        self.enqueued.push([method, time.time()])
        self.busy.write(True)
        # send Qt SIGNAL to address thread
        self.signal.emit(method, [args, kwargs])
        """
        self.queue.invokeMethod(self.driver,
                                'dequeue',
                                QtCore.Qt.QueuedConnection,
                                QtCore.Q_ARG('str', method),
                                QtCore.Q_ARG('list', [args, kwargs]))
                                """



"""GUI for displaying scans in progress, current slice etc."""

from PySide2 import QtCore, QtWidgets
import numpy as np

import WrightTools as wt
import yaqc_cmds
import yaqc_cmds.project.project_globals as g
import yaqc_cmds.sensors as sensors
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.classes as pc
import yaqc_cmds.somatic as somatic


class GUI(QtCore.QObject):
    def __init__(self):
        QtCore.QObject.__init__(self)
        self.create_frame()
        self.create_settings()
        self.on_sensors_changed()
        self.data = None

    def create_frame(self):
        self.main_widget = g.main_window.read().plot_widget
        # create main daq tab
        main_widget = self.main_widget
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 10, 0, 0)
        main_widget.setLayout(layout)
        # display
        # container widget
        display_container_widget = pw.ExpandingWidget()
        display_container_widget.setLayout(QtWidgets.QVBoxLayout())
        display_layout = display_container_widget.layout()
        display_layout.setMargin(0)
        layout.addWidget(display_container_widget)
        # big number
        big_number_container_widget = QtWidgets.QWidget()
        big_number_container_widget.setLayout(QtWidgets.QHBoxLayout())
        big_number_container_layout = big_number_container_widget.layout()
        big_number_container_layout.setMargin(0)
        self.big_display = pw.SpinboxAsDisplay(font_size=100)
        self.big_channel = pw.Label("channel", font_size=72)
        big_number_container_layout.addWidget(self.big_channel)
        big_number_container_layout.addStretch(1)
        big_number_container_layout.addWidget(self.big_display)
        display_layout.addWidget(big_number_container_widget)
        # plot
        self.plot_widget = pw.Plot1D()
        self.plot_scatter = self.plot_widget.add_scatter()
        self.plot_line = self.plot_widget.add_line()
        display_layout.addWidget(self.plot_widget)
        # vertical line
        line = pw.line("V")
        layout.addWidget(line)
        # settings
        settings_container_widget = QtWidgets.QWidget()
        settings_scroll_area = pw.scroll_area()
        settings_scroll_area.setWidget(settings_container_widget)
        settings_scroll_area.setMinimumWidth(300)
        settings_scroll_area.setMaximumWidth(300)
        settings_container_widget.setLayout(QtWidgets.QVBoxLayout())
        self.settings_layout = settings_container_widget.layout()
        self.settings_layout.setMargin(5)
        layout.addWidget(settings_scroll_area)
        g.shutdown.read().connect(self.on_shutdown)

    def create_settings(self):
        # display settings
        input_table = pw.InputTable()
        input_table.add("Display", None)
        self.channel = pc.Combo()
        input_table.add("Channel", self.channel)
        self.axis = pc.Combo()
        input_table.add("X-Axis", self.axis)
        self.axis_units = pc.Combo()
        input_table.add("X-Units", self.axis_units)
        self.settings_layout.addWidget(input_table)
        # global daq settings
        input_table = pw.InputTable()
        input_table.add("Settings", None)
        # input_table.add("ms Wait", ms_wait)
        for sensor in sensors.sensors:
            input_table.add(sensor.name, None)
            input_table.add("Status", sensor.busy)
            input_table.add("Freerun", sensor.freerun)
            input_table.add("Time", sensor.measure_time)
        input_table.add("Scan", None)
        # input_table.add("Loop Time", loop_time)
        self.idx_string = pc.String(initial_value="None", display=True)
        input_table.add("Scan Index", self.idx_string)
        self.settings_layout.addWidget(input_table)
        # stretch
        self.settings_layout.addStretch(1)

    def on_channels_changed(self):
        new = list(sensors.get_channels_dict())
        self.channel.set_allowed_values(new)

    def on_data_file_created(self):
        with somatic._wt5.data_container as data:
            allowed = [x.split()[0] for x in data.attrs["axes"]]
            if "wa" in allowed:
                allowed.remove("wa")
            self.axis.set_allowed_values(allowed)
            self.on_axis_updated()

    def on_axis_updated(self):
        with somatic._wt5.data_container as data:
            axis = data[self.axis.read()]
            units = axis.attrs.get("units")
            units = [units] + list(wt.units.get_valid_conversions(units))
            self.axis_units.set_allowed_values(units)

    def on_data_file_written(self):
        with somatic._wt5.data_container as data:
            last_idx_written = somatic._wt5.data_container.last_idx_written
            self.idx_string.write(str(last_idx_written))
            if data is None or last_idx_written is None:
                return
            # data
            x_units = self.axis_units.read()
            idx = last_idx_written
            axis = data[self.axis.read()]
            limits = list(
                wt.units.convert(
                    [np.nanmin(axis.full), np.nanmax(axis.full)], axis.attrs.get("units"), x_units
                )
            )
            channel = data[self.channel.read()]
            plot_idx = list(last_idx_written + (0,) * (channel.ndim - len(last_idx_written)))
            plot_idx[self.axis.read_index()] = slice(None)

            plot_idx = tuple(plot_idx)
            try:
                xi = wt.units.convert(
                    axis[wt.kit.valid_index(plot_idx, axis.shape)],
                    axis.attrs.get("units"),
                    x_units,
                )
                yi = channel[wt.kit.valid_index(plot_idx, channel.shape)]
                self.plot_scatter.setData(xi, yi)
            except (TypeError, ValueError) as e:
                print(e)
                pass
            # limits
            try:
                self.plot_widget.set_xlim(min(limits), max(limits))
                self.plot_widget.set_ylim(np.min(channel), np.max(channel))
            except Exception as e:
                print(e)
                pass

    def on_sensors_changed(self):
        for s in sensors.sensors:
            s.update_ui.connect(self.update_big_number)
        self.on_channels_changed()

    def on_shutdown(self):
        pass

    def stop(self):
        pass

    def update_big_number(self):
        channel = self.channel.read()
        if channel == "None":
            return
        sensor = sensors.get_channels_dict()[channel]
        num = sensor.channels[channel]
        if not np.isscalar(num):
            channel = f"max({channel})"
            num = np.max(num)
        self.big_channel.setText(channel)
        self.big_display.setValue(num)


gui = GUI()
somatic.signals.data_file_written.connect(gui.on_data_file_written)
somatic.signals.data_file_created.connect(gui.on_data_file_created)
sensors.signals.channels_changed.connect(gui.on_channels_changed)
sensors.signals.sensors_changed.connect(gui.on_sensors_changed)
gui.axis.updated.connect(gui.on_axis_updated)
gui.axis.updated.connect(gui.on_data_file_written)
gui.axis_units.updated.connect(gui.on_data_file_written)



import numpy as np


def process(destinations_list):
    # out
    out = [idx for idx in np.ndindex(destinations_list[0].arr.shape)]
    # generate slices list of dictionaries
    if len(destinations_list[-1].arr.shape) > 1:
        # multidimensional scan
        slices = []
        for i, idx in enumerate(np.ndindex(destinations_list[-1].arr.shape[:-1])):
            s = {}
            s["index"] = destinations_list[-1].arr.shape[-1] * i
            s["name"] = destinations_list[-1].hardware.name
            s["units"] = destinations_list[-1].units
            s["points"] = destinations_list[-1].arr[idx]
            if destinations_list[-1].method == "set_position":
                s["use actual"] = True
            else:
                s["use actual"] = False
            slices.append(s)
    else:
        # 1D scan
        s = {}
        s["index"] = 0
        s["name"] = destinations_list[0].hardware.name
        s["units"] = destinations_list[0].units
        s["points"] = destinations_list[0].arr
        if destinations_list[0].method == "set_position":
            s["use actual"] = True
        else:
            s["use actual"] = False
        slices = [s]
    return out, slices
